@@ -1,3 +1,35 @@
+/*
+ * DiskSim Storage Subsystem Simulation Environment (Version 4.0)
+ * Revision Authors: John Bucy, Greg Ganger
+ * Contributors: John Griffin, Jiri Schindler, Steve Schlosser
+ *
+ * Copyright (c) of Carnegie Mellon University, 2001-2008.
+ *
+ * This software is being provided by the copyright holders under the
+ * following license. By obtaining, using and/or copying this software,
+ * you agree that you have read, understood, and will comply with the
+ * following terms and conditions:
+ *
+ * Permission to reproduce, use, and prepare derivative works of this
+ * software is granted provided the copyright and "No Warranty" statements
+ * are included with all reproductions and derivative works and associated
+ * documentation. This software may also be redistributed without charge
+ * provided that the copyright and "No Warranty" statements are included
+ * in all redistributions.
+ *
+ * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
+ * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
+ * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
+ * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
+ * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
+ * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
+ * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
+ * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
+ * OR DOCUMENTATION.
+ *
+ */
+
+
 
 /*
  * DiskSim Storage Subsystem Simulation Environment (Version 2.0)
@@ -74,15 +106,29 @@
 #include "disksim_ioqueue.h"
 #include "disksim_cache.h"
 #include "disksim_bus.h"
+#include "config.h"
+
+#include "modules/modules.h"
 
 /* Currently, controllers can not communicate via ownership-type buses */
 
 
-controller * getctlr (int ctlrno)
+INLINE controller * getctlr (int ctlrno)
 {
    ctlrinfo_t *ctlrinfo = disksim->ctlrinfo;
    ASSERT1((ctlrno >= 0) && (ctlrno < ctlrinfo->numcontrollers), "ctlrno", ctlrno);
-   return (&ctlrinfo->controllers[ctlrno]);
+   return (ctlrinfo->controllers[ctlrno]);
+}
+
+controller *getctlrbyname(char *name, int *num) {
+  int c;
+  for(c = 0; c < disksim->ctlrinfo->numcontrollers; c++) {
+    if(!strcmp(name, disksim->ctlrinfo->controllers[c]->name)) {
+      if(num) *num = c;
+      return disksim->ctlrinfo->controllers[c];
+    }
+  }
+  return 0;
 }
 
 
@@ -163,7 +209,7 @@ fprintf (outputfile, "Must get ownership of bus\n");
       curr->time = (double) -1.0;
    } else {
       fprintf(stderr, "Wrong bus owned for transfer desired\n");
-      exit(0);
+      exit(1);
    }
 }
 
@@ -198,7 +244,7 @@ void controller_send_event_up_path (controller *currctlr, ioreq_event *curr, dou
       curr->time = (double) -1.0;
    } else {
       fprintf(stderr, "Wrong bus owned for transfer desired\n");
-      exit(0);
+      exit(1);
    }
 }
 
@@ -214,10 +260,10 @@ fprintf (outputfile, "Entered controller_get_data_transfered - ctlno %d, devno %
 */
    while (tmp) {
       if (tmp->devno == devno) {
-         tmptime = ((ioreq_event *) tmp->buf)->time;
+         tmptime = ((ioreq_event *) tmp->tempptr1)->time;
          tmptime = (tmptime - simtime) / tmp->time;
-         tmpblks = ((ioreq_event *) tmp->buf)->blkno - tmp->blkno;
-         tmpblks += ((ioreq_event *) tmp->buf)->bcount;
+         tmpblks = ((ioreq_event *) tmp->tempptr1)->blkno - tmp->blkno;
+         tmpblks += ((ioreq_event *) tmp->tempptr1)->bcount;
          return((int) ((double) tmpblks - tmptime));
       }
       tmp = tmp->next;
@@ -261,7 +307,7 @@ fprintf (outputfile, "Ownership granted for bus %d - ctlno %d\n", busno, ctlno);
                                   break;
       default:
                fprintf(stderr, "Unknown event type at controller_bus_ownership_grant - %d\n", tmp->type);
-               exit(0);
+               exit(1);
    }
    currctlr->waitingforbus += arbdelay;
    bus_delay(busno, CONTROLLER, currctlr->ctlno, tmp->time, tmp);
@@ -292,7 +338,7 @@ void controller_remove_type_from_buswait (controller *currctlr, int type)
    }
    if (tmp->time == -1.0) {
       fprintf(stderr, "In remove_type_from_buswait, the event is in transit\n");
-      exit(0);
+      exit(1);
    } else {
       busno = controller_get_downward_busno(currctlr, tmp, NULL);
       bus_remove_from_arbitration(busno, tmp);
@@ -342,7 +388,7 @@ fprintf (outputfile, "Controller_bus_delay_complete - ctlno %d, busno %d, blkno
       }
    } else {
       fprintf(stderr, "Non-matching busno for request in controller_bus_delay_complete: busno %d\n", busno);
-      exit(0);
+      exit(1);
    }
 }
 
@@ -372,288 +418,202 @@ void controller_event_arrive (int ctlno, ioreq_event *curr)
 
       default:
                fprintf(stderr, "Unknown controller type in controller_interrupt_arrive: %d\n", currctlr->type);
-               exit(0);
+               exit(1);
    }
 }
 
 
-void controller_read_toprints (FILE *parfile)
+
+
+int disksim_ctlr_stats_loadparams(struct lp_block *b) {
+
+  ctlrinfo_t *ctlrinfo;
+
+  if(disksim->ctlrinfo == NULL) {
+    disksim->ctlrinfo = malloc(sizeof(ctlrinfo_t));
+    if(!disksim->ctlrinfo) return 0;
+    bzero((char *)disksim->ctlrinfo, sizeof(ctlrinfo_t));
+  }
+  ctlrinfo = disksim->ctlrinfo;
+
+  //  #include "modules/disksim_ctlr_stats_param.c"
+  lp_loadparams(0, b, &disksim_ctlr_stats_mod);
+
+  return 1;
+}
+
+
+
+
+controller *controller_copy(controller *orig) {
+  controller *result = malloc(sizeof(controller));
+  if(!result) return 0;
+  memcpy(result, orig, sizeof(controller));
+  result->cache = orig->cache->cache_copy(orig->cache);
+  result->queue = ioqueue_copy(orig->queue);
+
+  /* don't copy these */
+  result->connections = 0;
+  result->datatransfers = 0;
+  result->hostwaiters = 0;
+  result->buswait = 0;
+  result->outbuses = 0;
+
+  return result;
+}
+
+struct controller *disksim_ctlr_loadparams(struct lp_block *b)
 {
-   ctlrinfo_t *ctlrinfo;
+  int c;
+  controller *result;
+  /* temp vars for parameters */
+  
 
-   if (disksim->ctlrinfo == NULL) {
-      disksim->ctlrinfo = DISKSIM_malloc (sizeof(ctlrinfo_t));
-      bzero ((char *)disksim->ctlrinfo, sizeof(ctlrinfo_t));
-   }
-   ctlrinfo = disksim->ctlrinfo;
 
-   getparam_bool(parfile, "Print controller cache stats?", &ctlrinfo->ctl_printcachestats);
-   getparam_bool(parfile, "Print controller size stats?", &ctlrinfo->ctl_printsizestats);
-   getparam_bool(parfile, "Print controller locality stats?", &ctlrinfo->ctl_printlocalitystats);
-   getparam_bool(parfile, "Print controller blocking stats?", &ctlrinfo->ctl_printblockingstats);
-   getparam_bool(parfile, "Print controller interference stats?", &ctlrinfo->ctl_printinterferestats);
-   getparam_bool(parfile, "Print controller queue stats?", &ctlrinfo->ctl_printqueuestats);
-   getparam_bool(parfile, "Print controller crit stats?", &ctlrinfo->ctl_printcritstats);
-   getparam_bool(parfile, "Print controller idle stats?", &ctlrinfo->ctl_printidlestats);
-   getparam_bool(parfile, "Print controller intarr stats?", &ctlrinfo->ctl_printintarrstats);
-   getparam_bool(parfile, "Print controller streak stats?", &ctlrinfo->ctl_printstreakstats);
-   getparam_bool(parfile, "Print controller stamp stats?", &ctlrinfo->ctl_printstampstats);
-   getparam_bool(parfile, "Print controller per-device stats?", &ctlrinfo->ctl_printperdiskstats);
+
+  /* initialize top-level controller info struct if necessary */
+  if (disksim->ctlrinfo == NULL) {
+    disksim->ctlrinfo = DISKSIM_malloc (sizeof(ctlrinfo_t));
+    bzero ((char *)disksim->ctlrinfo, sizeof(ctlrinfo_t));
+  }
+
+  
+  /* find a free slot in the controller table for the controller
+   * we're creating */
+  for(c = 0; c < disksim->ctlrinfo->ctlrs_len; c++) {
+    if(!disksim->ctlrinfo->controllers[c]) { goto foundslot; } 
+  }
+
+  /* didn't find a free slot -- resize the controller ptr array */
+  {
+    int newlen = c ? 2*c : 2;
+    int zerooff = (newlen == 2) ? 0 : c;
+    int zerolen = ((newlen == 2) ? 2 : (newlen / 2)) * sizeof(int *);
+    
+    disksim->ctlrinfo->controllers = realloc(disksim->ctlrinfo->controllers,
+					     newlen * sizeof(int *));
+    
+    bzero(disksim->ctlrinfo->controllers + zerooff, zerolen);
+    disksim->ctlrinfo->ctlrs_len = newlen;
+  }
+
+ foundslot:
+  
+  disksim->ctlrinfo->numcontrollers++;
+
+  /* allocate a new controller struct */
+  result = malloc(sizeof(controller));
+  if(!result) return 0;
+  bzero(result, sizeof(controller));
+
+  disksim->ctlrinfo->controllers[c] = result;
+  bzero(result, sizeof(controller));
+
+  result->name = strdup(b->name);
+    
+  //  #include "modules/disksim_ctlr_param.c"
+  lp_loadparams(result, b, &disksim_ctlr_mod);
+
+
+  /* Overhead for propogating request to lower-level component */
+  result->ovrhd_disk_request = 0.5;
+
+  /* Overhead for propogating ready-to-transfer to upper-level component */
+  result->ovrhd_ready = 0.5;
+
+  /* Overhead for propogating ready-to-transfer to lower-level component */
+  result->ovrhd_disk_ready = 0.5;
+
+  /* Overhead for propogating disconnect to upper-level component */
+  result->ovrhd_disconnect = 0.5;
+
+  /* Overhead for propogating disconnect to lower-level component */
+  result->ovrhd_disk_disconnect = 0.5;
+
+  /* Overhead for propogating reconnect to upper-level component */
+  result->ovrhd_reconnect = 0.5;
+
+  /* Overhead for propogating reconnect to lower-level component */
+  result->ovrhd_disk_reconnect = 0.5;
+
+  /* Overhead for propogating complete to upper-level component */
+  result->ovrhd_complete = 0.5;
+
+  /* Overhead for propogating complete to lower-level component */
+  result->ovrhd_disk_complete = 0.5;
+
+  /* Overhead for propogating disconnect/complete-complete to
+     upper-level component */
+  result->ovrhd_reset = 0.5;
+
+
+
+  
+  return result;
 }
 
 
-void controller_read_specs(FILE *parfile)
-{
-   int specno_expected = 1;
-   int specno;
-   int ctlno = 0;
-   int copies;
-   int maxqlen;
-   int i;
-   ctlrinfo_t *ctlrinfo;
-
-   if (disksim->ctlrinfo == NULL) {
-      disksim->ctlrinfo = DISKSIM_malloc (sizeof(ctlrinfo_t));
-      bzero ((char *)disksim->ctlrinfo, sizeof(ctlrinfo_t));
-   }
-   ctlrinfo = disksim->ctlrinfo;
-
-
-   getparam_int(parfile, "\nNumber of controllers", &ctlrinfo->numcontrollers, 1, 0, 0);
-
-   if (ctlrinfo->numcontrollers == 0) {
-      return;
-   }
-
-   
-   ctlrinfo->controllers = DISKSIM_malloc (ctlrinfo->numcontrollers * (sizeof(controller)));
-
-   while (ctlno < ctlrinfo->numcontrollers) {
-      controller *currctlr = getctlr(ctlno);
-      if (fscanf(parfile, "\nController Spec #%d\n", &specno) != 1) {
-         fprintf(stderr, "Error reading controller spec number\n");
-         exit(0);
-      }
-      if (specno != specno_expected) {
-         fprintf(stderr, "Unexpected value for controller spec number: specno %d, expected %d\n", specno, specno_expected);
-         exit(0);
-      }
-      fprintf (outputfile, "\nController Spec #%d\n", specno);
-      specno_expected++;
-
-      getparam_int(parfile, "# controllers with Spec", &copies, 1, 0, 0);
-
-      if (copies == 0) {
-         copies = ctlrinfo->numcontrollers - ctlno;
-      }
-
-      if ((ctlno + copies) > ctlrinfo->numcontrollers) {
-         fprintf(stderr, "Too many controller specifications provided\n");
-         exit(0);
-      }
-
-      getparam_int(parfile, "Controller Type", &currctlr->type, 3, MINCTLTYPE, MAXCTLTYPE);
-      getparam_double(parfile, "Scale for delays", &currctlr->timescale, 1, (double) 0.0, (double) 0.0);
-      getparam_double(parfile, "Bulk sector transfer time", &currctlr->blktranstime, 1, (double) 0.0, (double) 0.0);
-
-	/* Overhead for propogating request to lower-level component */
-      currctlr->ovrhd_disk_request = 0.5;
-	/* Overhead for propogating ready-to-transfer to upper-level component */
-      currctlr->ovrhd_ready = 0.5;
-	/* Overhead for propogating ready-to-transfer to lower-level component */
-      currctlr->ovrhd_disk_ready = 0.5;
-	/* Overhead for propogating disconnect to upper-level component */
-      currctlr->ovrhd_disconnect = 0.5;
-	/* Overhead for propogating disconnect to lower-level component */
-      currctlr->ovrhd_disk_disconnect = 0.5;
-	/* Overhead for propogating reconnect to upper-level component */
-      currctlr->ovrhd_reconnect = 0.5;
-	/* Overhead for propogating reconnect to lower-level component */
-      currctlr->ovrhd_disk_reconnect = 0.5;
-	/* Overhead for propogating complete to upper-level component */
-      currctlr->ovrhd_complete = 0.5;
-	/* Overhead for propogating complete to lower-level component */
-      currctlr->ovrhd_disk_complete = 0.5;
-	/* Overhead for propogating disconnect/complete-complete to upper-level component */
-      currctlr->ovrhd_reset = 0.5;
-
-      getparam_int(parfile, "Maximum queue length", &maxqlen, 1, 0, 0);
-      currctlr->maxoutstanding = maxqlen + 1;
-      getparam_int(parfile, "Print stats for controller", &currctlr->printstats, 3, 0, 1);
-
-      for (i=ctlno; i<(ctlno+copies); i++) {
-         controller *currctlr2 = getctlr(i);
-	 currctlr2->type = currctlr->type;
-	 currctlr2->timescale = currctlr->timescale;
-	 currctlr2->blktranstime = currctlr->blktranstime;
-         currctlr2->ovrhd_disk_request = currctlr->ovrhd_disk_request;
-         currctlr2->ovrhd_ready = currctlr->ovrhd_ready;
-         currctlr2->ovrhd_disk_ready = currctlr->ovrhd_disk_ready;
-         currctlr2->ovrhd_disconnect = currctlr->ovrhd_disconnect;
-         currctlr2->ovrhd_disk_disconnect = currctlr->ovrhd_disk_disconnect;
-         currctlr2->ovrhd_reconnect = currctlr->ovrhd_reconnect;
-         currctlr2->ovrhd_disk_reconnect = currctlr->ovrhd_disk_reconnect;
-         currctlr2->ovrhd_complete = currctlr->ovrhd_complete;
-         currctlr2->ovrhd_disk_complete = currctlr->ovrhd_disk_complete;
-         currctlr2->ovrhd_reset = currctlr->ovrhd_reset;
-	 currctlr2->maxoutstanding = currctlr->maxoutstanding;
-	 currctlr2->printstats = currctlr->printstats;
-	 currctlr2->buswait = NULL;
-	 currctlr2->connections = NULL;
-	 currctlr2->datatransfers = NULL;
-	 currctlr2->hostwaiters = NULL;
-	 currctlr2->numinbuses = 0;
-      }
-      if (currctlr->type == CTLR_SMART) {
-	 controller_smart_read_specs(parfile, ctlrinfo->controllers, ctlno, copies);
-      }
-      ctlno += copies;
-   }
-}
-
-
-void controller_read_physorg (FILE *parfile)
-{
-   ctlrinfo_t *ctlrinfo = disksim->ctlrinfo;
-   int i = 0;
-   int usectlnum = 0;
-   int j;
-   int ctlno;
-   char tmp[20];
-   char tmpvals[20];
-   int numrets = 0;
-   int outbuscnt;
-   int outbusadd;
-   char outbuses[20];
-   char op;
-   int val;
-
-   while (i < ctlrinfo->numcontrollers) {
-
-      if (fscanf(parfile, "\nController #%s\n", tmpvals) != 1) {
-         fprintf(stderr, "Error reading controller portion of physical organization\n");
-         exit(0);
-      }
-      if ((numrets = sscanf(tmpvals, "%d%s", &ctlno, tmp)) < 1) {
-         fprintf(stderr, "Error parsing controller portion of physical organization\n");
-         exit(0);
-      } else if (numrets == 1) {
-         /* sprintf(tmp, ""); */
-         tmp[0] = (char) 0;
-      }
-      if (ctlno != (i + 1)) {
-         fprintf(stderr, "Controller numbers in physical organization appear out of order\n");
-         exit(0);
-      }
-      fprintf (outputfile, "\nController #%d%s\n", ctlno, tmp);
-
-      j = ctlno;
-      if (strcmp(tmp, "") != 0) {
-         if (sscanf(tmp, "%c%d", &op, &val) != 2) {
-            fprintf(stderr, "Can't parse value for 'Controller #' %d in physical organization\n", ctlno);
-            exit(0);
-         }
-         if ((op != '-') || (val <= ctlno) || (val > ctlrinfo->numcontrollers)) {
-            fprintf(stderr, "Illegal value for 'Controller #' %d in physical organization - %d%c%d\n", ctlno, ctlno, op, val);
-            exit(0);
-         }
-         j += val - ctlno;
-      }
-
-      getparam_int(parfile, "# of connected buses", &outbuscnt, 3, 1, MAXOUTBUSES);
-
-      if (fscanf(parfile, "Connected buses: %s\n", outbuses) != 1) {
-         fprintf(stderr, "Error reading 'Connected buses' for controller %d\n", ctlno);
-         exit(0);
-      }
-      if ((outbuses[0] != '#') && ((outbuses[0] > '9') || (outbuses[0] < '1'))) {
-         fprintf(stderr, "Invalid value for 'Connected buses' for controller %d", ctlno);
-         exit(0);
-      }
-      fprintf (outputfile, "Connected buses:   %s\n", outbuses);
-
-      usectlnum = 0;
-      if (outbuses[0] == '#') {
-         usectlnum = 1;
-         outbusadd = 0;
-         if (strcmp(outbuses, "#") != 0) {
-            if (sscanf(outbuses, "#%c%d", &op, &val) != 2) {
-               fprintf(stderr, "Invalid value for 'Connected buses' for controller %d", ctlno);
-               exit(0);
-            }
-            if (((op != '+') && (op != '-')) || (val < 0)) {
-               fprintf(stderr, "Invalid value for 'Connected buses' for controller %d", ctlno);
-               exit(0);
-            }
-            if (op == '+') {
-               outbusadd = val;
-            } else {
-               outbusadd = -val;
-            }
-         }
-      } else {
-         sscanf(outbuses, "%d", &val);
-      }
-
-      for (; i < j; i++) {
-         controller *currctlr = getctlr(i);
-	 currctlr->numoutbuses = outbuscnt;
-         if (usectlnum) {
-            currctlr->outbuses[0] = i + outbusadd;
-         } else {
-            currctlr->outbuses[0] = val - 1;
-         }
-      }
-   }
-}
 
 
 void controller_read_logorg (FILE *parfile)
 {
-   ctlrinfo_t *ctlrinfo = disksim->ctlrinfo;
+  assert(0);
+/*     ctlrinfo_t *ctlrinfo = disksim->ctlrinfo; */
 
-   getparam_int(parfile, "\n# of controller-level organizations", &ctlrinfo->numctlorgs, 3, 0, MAXLOGORGS);
-   fscanf(parfile, "\n");
+/*     getparam_int(parfile, "\n# of controller-level organizations", &ctlrinfo->numctlorgs, 3, 0, MAXLOGORGS); */
+/*     fscanf(parfile, "\n"); */
 
-   if (ctlrinfo->numctlorgs > 0) {
-      fprintf (stderr, "Controller-level logical organizations are not currently supported\n");
-      exit (0);
-      ctlrinfo->ctlorgs = logorg_readparams(parfile, ctlrinfo->numctlorgs,
-             ctlrinfo->ctl_printlocalitystats, ctlrinfo->ctl_printblockingstats,
-             ctlrinfo->ctl_printinterferestats, ctlrinfo->ctl_printstreakstats,
-             ctlrinfo->ctl_printstampstats, ctlrinfo->ctl_printintarrstats,
-             ctlrinfo->ctl_printidlestats, ctlrinfo->ctl_printsizestats);
-   }
+/*     if (ctlrinfo->numctlorgs > 0) { */
+/*        fprintf (stderr, "Controller-level logical organizations are not currently supported\n"); */
+/*        exit (0); */
+/*        ctlrinfo->ctlorgs = logorg_readparams(parfile, ctlrinfo->numctlorgs, */
+/*               ctlrinfo->ctl_printlocalitystats, ctlrinfo->ctl_printblockingstats, */
+/*               ctlrinfo->ctl_printinterferestats, ctlrinfo->ctl_printstreakstats, */
+/*               ctlrinfo->ctl_printstampstats, ctlrinfo->ctl_printintarrstats, */
+/*               ctlrinfo->ctl_printidlestats, ctlrinfo->ctl_printsizestats); */
+/*     } */
 }
 
 
-int controller_set_depth (int ctlno, int inbusno, int depth, int slotno)
+int 
+controller_set_depth (int ctlno, 
+		      int inbusno, 
+		      int depth, 
+		      int slotno)
 {
    controller *currctlr = getctlr (ctlno);
    intchar ret;
    int i;
    int cnt;
+
 /*
 fprintf (outputfile, "controller_set_depth %d, inbusno %d, depth %d, slotno %d\n", ctlno, inbusno, depth, slotno);
 */
+
    cnt = currctlr->numinbuses;
    if ((cnt > 0) && (currctlr->depth[(cnt-1)] < depth)) {
-      return(0);
+      return 0;
    }
    currctlr->numinbuses++;
+
    if ((cnt + 1) > MAXINBUSES) {
-      fprintf(stderr, "Too many inbuses specified for controller %d : %d\n", ctlno, (cnt+1));
-      exit(0);
+      fprintf(stderr, 
+	      "*** Too many inbuses specified for controller %d : %d (check BUG #30)\n", 
+	      ctlno, (cnt+1));
+      ddbg_assert(0);
    }
+
    currctlr->inbuses[cnt] = inbusno;
    currctlr->depth[cnt] = depth;
    currctlr->slotno[cnt] = slotno;
    ret.value = 0;
+
    for (i = 0; i < currctlr->numoutbuses; i++) {
       ret.byte[i] = (char) currctlr->outbuses[i];
       currctlr->outslot[i] = bus_get_controller_slot(ret.byte[i], ctlno);
    }
-   return(ret.value);
+
+   return ret.value;
 }
 
 
@@ -687,7 +647,7 @@ int controller_get_outslot (int ctlno, int busno)
          break;
       }
    }
-   ASSERT(slotno != 0);
+/*     ASSERT(slotno != 0); */
    return(slotno);
 }
 
@@ -727,43 +687,6 @@ int controller_get_maxoutstanding (int ctlno)
 }
 
 
-void controller_param_override (char *paramname, char *paramval, int first, int last)
-{
-   ctlrinfo_t *ctlrinfo = disksim->ctlrinfo;
-   int i;
-
-   if (first == -1) {
-      first = 0;
-      last = ctlrinfo->numcontrollers - 1;
-   } else if ((first < 0) || (last >= ctlrinfo->numcontrollers) || (last < first)) {
-      fprintf(stderr, "Invalid range at controller_param_override: %d - %d\n", first, last);
-      exit(0);
-   }
-   for (i=first; i<=last; i++) {
-      controller *currctlr = getctlr(i);
-      if ((strcmp(paramname, "ioqueue_schedalg") == 0) || (strcmp(paramname, "ioqueue_seqscheme") == 0) || (strcmp(paramname, "ioqueue_cylmaptype") == 0) || (strcmp(paramname, "ioqueue_to_time") == 0) || (strcmp(paramname, "ioqueue_timeout_schedalg") == 0) || (strcmp(paramname, "ioqueue_priority_schedalg") == 0) || (strcmp(paramname, "ioqueue_priority_mix") == 0)) {
-         ioqueue_param_override(currctlr->queue, paramname, paramval);
-      } else if ((strcmp(paramname, "cache_size") == 0) || (strcmp(paramname, "cache_segcount") == 0) || (strcmp(paramname, "cache_linesize") == 0) || (strcmp(paramname, "cache_bitgran") == 0) || (strcmp(paramname, "cache_lockgran") == 0) || (strcmp(paramname, "cache_readshare") == 0) || (strcmp(paramname, "cache_maxreqsize") == 0) || (strcmp(paramname, "cache_replace") == 0) || (strcmp(paramname, "cache_writescheme") == 0) || (strcmp(paramname, "cache_readprefetch") == 0) || (strcmp(paramname, "cache_writeprefetch") == 0) || (strcmp(paramname, "cache_linebyline") == 0)) {
-         cache_param_override(currctlr->cache, paramname, paramval);
-      } else if (strcmp(paramname, "maxqlen") == 0) {
-         int maxqlen;
-         if (sscanf(paramval, "%d\n", &maxqlen) != 1) {
-            fprintf(stderr, "Error reading usequeue in iodriver_param_override\n");
-            exit(0);
-         }
-         if (maxqlen < 0) {
-            fprintf(stderr, "Invalid value for maxqlen in controller_param_override: %d\n", maxqlen);
-            exit(0);
-         }
-         currctlr->maxoutstanding = maxqlen + 1;
-      } else {
-         fprintf(stderr, "Unsupported param to override at controller_param_override: %s\n", paramname);
-         exit(0);
-      }
-   }
-}
-
-
 void controller_setcallbacks ()
 {
    controller_smart_setcallbacks ();
@@ -843,7 +766,7 @@ void controller_printstats()
 
          default:
             fprintf(stderr, "Unknown controller type in controller_printstats: %d\n", currctlr->type);
-            exit(0);
+            exit(1);
        }
    }
    fprintf (outputfile, "Total controller bus wait time: %f\n", waitingforbus);
@@ -854,3 +777,58 @@ void controller_cleanstats()
 {
 }
 
+
+int load_ctlr_topo(struct lp_topospec *t, int *inbus) {
+  int c;
+  controller *ctlr;
+  int num;
+  struct lp_topospec *ts;
+  int slots;
+  assert(!strcmp(t->type, disksim_mods[DISKSIM_MOD_CTLR]->name));
+
+  ctlr = getctlrbyname(t->name, &num);
+  if(!ctlr) {
+    fprintf(stderr, "*** error: no such controller %s\n", t->name);
+    return 0;
+  }
+
+
+  if(inbus) {
+    ctlr->inbuses[0] = *inbus;
+  }
+
+  if(!ctlr->outbuses) {
+    ctlr->numoutbuses = 0;
+    slots = t->l->values_len;
+    ctlr->outbuses = malloc(slots * sizeof(int));
+    bzero(ctlr->outbuses, slots * sizeof(int));
+  }
+
+  for(c = 0; c < t->l->values_len; c++) {
+    if(!t->l->values[c]) continue;
+    assert(t->l->values[c]->v.t.len == 1);
+    ts = &t->l->values[c]->v.t.l[0];
+    if(!strcmp(ts->type, disksim_mods[DISKSIM_MOD_BUS]->name)) {
+      getbusbyname(ts->name, &ctlr->outbuses[c]);
+
+      if(ctlr->outbuses[c] == -1) {
+	fprintf(stderr, "*** error: No such bus: %s\n", ts->name);
+	return 0;
+      }
+
+      if(!load_bus_topo(ts, &num)) {
+	fprintf(stderr, "*** error: failed to load bus topology spec for controller %s.\n", t->name);
+	return 0;
+      }
+
+      ctlr->numoutbuses++;
+    }
+    else {
+      fprintf(stderr, "*** error: can only attach buses to controllers %s\n", t->name);
+      return 0;
+    }
+  }
+
+  return 1;
+}
+
