@@ -1,3 +1,35 @@
+/*
+ * DiskSim Storage Subsystem Simulation Environment (Version 4.0)
+ * Revision Authors: John Bucy, Greg Ganger
+ * Contributors: John Griffin, Jiri Schindler, Steve Schlosser
+ *
+ * Copyright (c) of Carnegie Mellon University, 2001-2008.
+ *
+ * This software is being provided by the copyright holders under the
+ * following license. By obtaining, using and/or copying this software,
+ * you agree that you have read, understood, and will comply with the
+ * following terms and conditions:
+ *
+ * Permission to reproduce, use, and prepare derivative works of this
+ * software is granted provided the copyright and "No Warranty" statements
+ * are included with all reproductions and derivative works and associated
+ * documentation. This software may also be redistributed without charge
+ * provided that the copyright and "No Warranty" statements are included
+ * in all redistributions.
+ *
+ * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
+ * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
+ * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
+ * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
+ * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
+ * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
+ * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
+ * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
+ * OR DOCUMENTATION.
+ *
+ */
+
+
 
 /*
  * DiskSim Storage Subsystem Simulation Environment (Version 2.0)
@@ -402,7 +434,7 @@ int disk_buffer_block_available (disk *currdisk, segment *seg, int blkno)
       }
    } else {
       fprintf(stderr, "Not actively using disk at disk_buffer_block_available\n");
-      exit(0);
+      exit(1);
    }
    return(TRUE);
 }
@@ -469,7 +501,7 @@ static void disk_buffer_stats (disk *currdisk, ioreq_event *curr, segment *seg,
             size = (double) (curr->blkno + curr->bcount - seg->startblkno);
 */
             fprintf(stderr, "Tail hits not currently allowed.\n");
-            exit(0);
+            exit(1);
          }
          if (seg->state == BUFFER_READING) {
 /*
@@ -490,7 +522,7 @@ static void disk_buffer_stats (disk *currdisk, ioreq_event *curr, segment *seg,
 
       default:
 	 fprintf(stderr, "Invalid hittype in disk_buffer_stats - blkno %d, bcount %d, state %d\n", curr->blkno, curr->bcount, seg->state);
-         exit(0);
+         exit(1);
    }
    return;
 }
@@ -545,14 +577,15 @@ static void disk_buffer_stats (disk *currdisk, ioreq_event *curr, segment *seg,
  * request. If it doesn't find one, it sets seg to NULL and hittype to
  * BUFFER_NOMATCH */
 
-static segment* disk_buffer_select_read_segment (disk *currdisk, diskreq *currdiskreq)
+static segment * 
+disk_buffer_select_read_segment(disk *currdisk, diskreq *currdiskreq)
 {
    segment *seg;
    ioreq_event *first_ioreq = currdiskreq->ioreqlist;
    ioreq_event *tmpioreq;
    int best_value = 0;
    int curr_value;
-   int curr_hittype;
+   int curr_hittype = 0; 
 
 if (disk_printhack && (simtime >= disk_printhacktime)) {
 fprintf (outputfile, "%12.6f  %8p  Entering disk_buffer_select_read_segment\n",simtime,currdiskreq);
@@ -564,109 +597,130 @@ fflush(outputfile);
    seg = currdisk->seglist;
    while (seg) {
       curr_value = -1;
-      if (currdiskreq->hittype == BUFFER_COLLISION) {
-      } else if (seg->recyclereq) {
-      } else if ((currdisk->dedicatedwriteseg) && 
-	  (seg == currdisk->dedicatedwriteseg)) {
-         /* check for collision with dirty data */
-         if ((seg->state == BUFFER_DIRTY) || (seg->state == BUFFER_WRITING)) {
-	    tmpioreq = currdiskreq->ioreqlist;
-	    while (tmpioreq) {
-	       if (disk_buffer_overlap(seg,tmpioreq)) {
-		  if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
+      if (currdiskreq->hittype == BUFFER_COLLISION) { } 
+      else if (seg->recyclereq) { } 
+      else if ((currdisk->dedicatedwriteseg) 
+	       && (seg == currdisk->dedicatedwriteseg)) 
+           {
+	     /* check for collision with dirty data */
+	     if ((seg->state == BUFFER_DIRTY) 
+		 || (seg->state == BUFFER_WRITING)) 
+	     {
+	       tmpioreq = currdiskreq->ioreqlist;
+	       while (tmpioreq) {
+		 if (disk_buffer_overlap(seg,tmpioreq)) {
+		   if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
 		     currdiskreq->hittype = BUFFER_COLLISION;
 		     currdiskreq->seg = NULL;
-		  }
-                  break;
+		   }
+		   break;
+		 }
+		 tmpioreq = tmpioreq->next;
 	       }
-	       tmpioreq = tmpioreq->next;
-	    }
-	 }
-      } else if (seg->state == BUFFER_EMPTY) {
-	 curr_value = 3;
-         curr_hittype = BUFFER_NOMATCH;
-      } else if ((first_ioreq->blkno < seg->startblkno) || 
-		 (first_ioreq->blkno >= seg->endblkno)) {
-	 curr_hittype = BUFFER_NOMATCH;
-	 if ((seg->state == BUFFER_CLEAN) && !(seg->diskreqlist)) {
-	    curr_value = 2;
-         } else if (seg->state == BUFFER_READING) {
-	    if ((seg->endblkno == first_ioreq->blkno) && 
-		currdisk->almostreadhits) {
-	       /* special case:  block currently being prefetched is
-		  the first block of this request.  We count this as
-		  a partial read hit.
-               */
-	       curr_hittype = BUFFER_PARTIAL;
-	       curr_value = 5;
-	    } else {
-	       if (!seg->diskreqlist->ioreqlist) {
-	          curr_value = 1;
-	       }
-            }
-         } else if ((first_ioreq->blkno < seg->startblkno) &&
-		    ((seg->state == BUFFER_DIRTY) || 
-		     (seg->state == BUFFER_WRITING))) {
-            /* check for collision with dirty data */
-	    tmpioreq = currdiskreq->ioreqlist;
-	    while (tmpioreq) {
-	       if (disk_buffer_overlap(seg,tmpioreq)) {
-		  if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
-		     currdiskreq->hittype = BUFFER_COLLISION;
-		     currdiskreq->seg = NULL;
-		  }
-		  break;
-	       }
-	       tmpioreq = tmpioreq->next;
-	    }
-	 }
-      } else {
-	 switch (seg->state) {
-	    case BUFFER_DIRTY:
-	       if (currdisk->enablecache) {
-	          if ((currdisk->readhitsonwritedata) &&
-		      (best_value < 10)) {
-		     curr_value = 11;
-	          } else {
-		     if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
-		        currdiskreq->hittype = BUFFER_COLLISION;
-		        currdiskreq->seg = NULL;
-             	     }
-	          }
-	       }
-	       break;
-	    case BUFFER_WRITING:
-	       if (currdisk->enablecache) {
-	          if ((currdisk->readhitsonwritedata) &&
-		      (best_value < 10)) {
-		     curr_value = 10;
-	          } else {
-		     if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
-		        currdiskreq->hittype = BUFFER_COLLISION;
-		        currdiskreq->seg = NULL;
-		     }
-	          }
-	       }
-	       break;
-	    case BUFFER_READING:
-	       if (currdisk->enablecache) {
-	          curr_value = 9;
-	       }
-	       break;
-	    case BUFFER_CLEAN:
-	      if (currdisk->readhitsonwritedata ||
-		  (seg->access && (seg->access->flags & READ))) {
-	         curr_value = 8;
-	      }
-	       break;
-	 }
-	 if (seg->endblkno >= (first_ioreq->blkno + first_ioreq->bcount)) {
-            curr_hittype = BUFFER_WHOLE;
-         } else {
+	     }
+	   } 
+      else if (seg->state == BUFFER_EMPTY) {
+	curr_value = 3;
+	curr_hittype = BUFFER_NOMATCH;
+      } 
+      else if ((first_ioreq->blkno < seg->startblkno) 
+	       || (first_ioreq->blkno >= seg->endblkno)) 
+      {
+	curr_hittype = BUFFER_NOMATCH;
+	if ((seg->state == BUFFER_CLEAN) && !(seg->diskreqlist)) {
+	  curr_value = 2;
+	} 
+	else if (seg->state == BUFFER_READING) {
+	  if ((seg->endblkno == first_ioreq->blkno) && 
+	      currdisk->almostreadhits) {
+	    /* special case: block currently being prefetched is the
+	     * first block of this request.  We count this as a
+	     * partial read hit.  
+	     */
 	    curr_hittype = BUFFER_PARTIAL;
-	    curr_value -= 4;
-	 }
+	    curr_value = 5;
+	  } 
+	  else {
+	    if (!seg->diskreqlist->ioreqlist) {
+	      curr_value = 1;
+	    }
+	  }
+	} 
+	else if ((first_ioreq->blkno < seg->startblkno) 
+		 && ((seg->state == BUFFER_DIRTY) 
+		     || (seg->state == BUFFER_WRITING))) 
+	{
+	  /* check for collision with dirty data */
+	  tmpioreq = currdiskreq->ioreqlist;
+	  while (tmpioreq) {
+	    if (disk_buffer_overlap(seg,tmpioreq)) {
+	      if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
+		currdiskreq->hittype = BUFFER_COLLISION;
+		currdiskreq->seg = NULL;
+	      }
+	      break;
+	    }
+	    tmpioreq = tmpioreq->next;
+	  }
+	}
+      } 
+      else {
+	switch (seg->state) {
+	case BUFFER_DIRTY:
+	  if (currdisk->enablecache) {
+	    if ((currdisk->readhitsonwritedata) 
+		&& (best_value < 10)) {
+	      curr_value = 11;
+	    } 
+	    else {
+	      if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
+		currdiskreq->hittype = BUFFER_COLLISION;
+		currdiskreq->seg = NULL;
+	      }
+	    }
+	  }
+	  break;
+	case BUFFER_WRITING:
+	  if (currdisk->enablecache) {
+	    if ((currdisk->readhitsonwritedata) 
+		&& (best_value < 10)) 
+            {
+	      curr_value = 10;
+	    } 
+	    else {
+	      if (!disk_buffer_reusable_segment_check(currdisk, seg)) {
+		currdiskreq->hittype = BUFFER_COLLISION;
+		currdiskreq->seg = NULL;
+	      }
+	    }
+	  }
+	  break;
+	case BUFFER_READING:
+	  if (currdisk->enablecache) {
+	    curr_value = 9;
+	  }
+	  break;
+	case BUFFER_CLEAN:
+	  if (currdisk->readhitsonwritedata 
+	      || (seg->access && (seg->access->flags & READ))) 
+	  {
+	    curr_value = 8;
+	  }
+	  break;
+
+	 default:
+	   ddbg_assert(0);
+	}
+
+	if (seg->endblkno >= (first_ioreq->blkno + first_ioreq->bcount)) {
+	  curr_hittype = BUFFER_WHOLE;
+	} 
+	else {
+	  curr_hittype = BUFFER_PARTIAL;
+	  curr_value -= 4;
+	}
       }
+
       if (curr_value > best_value) {
 	 currdiskreq->seg = seg;
 	 currdiskreq->hittype = curr_hittype;
@@ -676,20 +730,21 @@ fflush(outputfile);
    }
 
    /* If BUFFER_NOMATCH && BUFFER_READING, perform preemption check.
-      Note that currdiskreq->seg should be set before the call.
-   */
+    * Note that currdiskreq->seg should be set before the call.  
+    */
 
-   if ((best_value == 1) &&
-       !disk_buffer_stopable_access(currdisk,currdiskreq)) {
-      currdiskreq->seg = NULL;
+   if ((best_value == 1) 
+       && !disk_buffer_stopable_access(currdisk,currdiskreq)) 
+   {
+     currdiskreq->seg = NULL;
    }
-
+   
 if (disk_printhack && (simtime >= disk_printhacktime)) {
 fprintf (outputfile, "                        segment = %8p, hittype = %d\n",currdiskreq->seg,currdiskreq->hittype);
 fflush(outputfile);
 }
 
-   return(currdiskreq->seg);
+ return currdiskreq->seg;
 }
 
 
@@ -963,19 +1018,20 @@ segment * disk_buffer_select_segment (disk *currdisk, diskreq *currdiskreq,
 {
    segment *seg;
 
-   if (currdisk->acctime >= 0.0) {
+   if (currdisk->const_acctime) {
       currdiskreq->seg = NULL;
       currdiskreq->hittype = BUFFER_NOMATCH;
       return(NULL);
    }
 
-   if (currdiskreq->ioreqlist->flags & READ) {
+   if(currdiskreq->ioreqlist->flags & READ) {
       seg = disk_buffer_select_read_segment(currdisk, currdiskreq);
       if (currdisk->dedicatedwriteseg && (seg == currdisk->dedicatedwriteseg)) {
 	 fprintf(stderr, "Read request about to use the write segment\n");
-	 exit(0);
+	 exit(1);
       }
-   } else {			/* WRITE */
+   } 
+   else {			/* WRITE */
       seg = disk_buffer_select_write_segment(currdisk, currdiskreq);
    }
 
@@ -983,13 +1039,18 @@ segment * disk_buffer_select_segment (disk *currdisk, diskreq *currdiskreq,
    /* and this is a write following a non-sequential write, set the        */
    /* EXTRA_WRITE_DISCONNECT diskreq flag.  */
 
-   if (set_extradisc && currdisk->extra_write_disconnect && 
-       !currdisk->neverdisconnect && !currdisk->extradisc_diskreq && 
-       !(currdiskreq->ioreqlist->flags & READ) && seg && 
-       ((LRU_at_seg_list_head && !seg->prev) ||
-        (!LRU_at_seg_list_head && !seg->next)) &&
-       seg->access && !(seg->access->flags & READ) && 
-       (currdiskreq->ioreqlist->blkno != seg->endblkno)) {
+   if (set_extradisc 
+       && currdisk->extra_write_disconnect 
+       && !currdisk->neverdisconnect 
+       && !currdisk->extradisc_diskreq 
+       && !(currdiskreq->ioreqlist->flags & READ) 
+       && seg 
+       && ((LRU_at_seg_list_head && !seg->prev) 
+	   || (!LRU_at_seg_list_head && !seg->next)) 
+       && seg->access 
+       && !(seg->access->flags & READ) 
+       && (currdiskreq->ioreqlist->blkno != seg->endblkno)) 
+   {
       currdiskreq->flags |= EXTRA_WRITE_DISCONNECT;
    }
    return(seg);
@@ -1050,23 +1111,25 @@ fflush(outputfile);
 	 }
       } else {
 	 switch (seg->state) {
-	    case BUFFER_DIRTY:
-	    case BUFFER_WRITING:
-	       if (read_hit_on_dirty_data ||
-		   (!currdisk->readhitsonwritedata &&
-	            !disk_buffer_reusable_segment_check(currdisk, seg))) {
-	          return(BUFFER_COLLISION);
-	       }
-	       read_hit_on_dirty_data = TRUE;
-	       break;
-	    case BUFFER_READING:
-	       break;
-	    case BUFFER_CLEAN:
+	 case BUFFER_DIRTY:
+	 case BUFFER_WRITING:
+	   if (read_hit_on_dirty_data ||
+	       (!currdisk->readhitsonwritedata &&
+		!disk_buffer_reusable_segment_check(currdisk, seg))) {
+	     return(BUFFER_COLLISION);
+	   }
+	   read_hit_on_dirty_data = TRUE;
+	   break;
+	 case BUFFER_READING:
+	   break;
+	 case BUFFER_CLEAN:
 	       if (!currdisk->readhitsonwritedata &&
-		  seg->access && !(seg->access->flags & READ)) {
+		   seg->access && !(seg->access->flags & READ)) {
 		 continue;
 	       }
 	       break;
+	 default:
+	   ddbg_assert(0);
 	 }
 	 if (seg->endblkno >= (currioreq->blkno + currioreq->bcount)) {
             return_hittype = BUFFER_WHOLE;
@@ -1112,10 +1175,13 @@ fflush(outputfile);
    seg = currdisk->seglist;
    while (seg) {
       if (seg->recyclereq) {
-      } else if (seg->state == BUFFER_EMPTY) {
-      } else if (seg->state == BUFFER_READING) {
+      } 
+      else if (seg->state == BUFFER_EMPTY) {
+      } 
+      else if (seg->state == BUFFER_READING) {
 	 return(BUFFER_NOMATCH);
-      } else if ((currdisk->dedicatedwriteseg) && 
+      } 
+      else if ((currdisk->dedicatedwriteseg) && 
 	  (seg != currdisk->dedicatedwriteseg)) {
          /* check for collision with uncleanable data */
          if ((seg->state == BUFFER_CLEAN) && seg->diskreqlist) {
@@ -1124,21 +1190,26 @@ fflush(outputfile);
 	       return(BUFFER_COLLISION);
 	    }
 	 }
-      } else if ((seg->state == BUFFER_CLEAN) || (seg->state == BUFFER_DIRTY)) {
-         /* check for collision with uncleanable data */
-	 if (seg->diskreqlist) {
-	       if (disk_buffer_overlap(seg,currioreq) &&
-		   !disk_buffer_reusable_segment_check(currdisk, seg)) {
-	          return(BUFFER_COLLISION);
-	       }
-	 }
-      } else if (seg->outstate == BUFFER_TRANSFERING) {
+      } 
+      else if ((seg->state == BUFFER_CLEAN) || 
+	       (seg->state == BUFFER_DIRTY)) 
+	{
+	  /* check for collision with uncleanable data */
+	  if (seg->diskreqlist) {
+	    if (disk_buffer_overlap(seg,currioreq) &&
+		!disk_buffer_reusable_segment_check(currdisk, seg)) {
+	      return(BUFFER_COLLISION);
+	    }
+	  }
+      } 
+      else if (seg->outstate == BUFFER_TRANSFERING) {
          /* check for collision with uncleanable data */
 	 if (disk_buffer_overlap(seg,currioreq) &&
 	     !disk_buffer_reusable_segment_check(currdisk, seg)) {
 	    return(BUFFER_COLLISION);
 	 }
-      } else {
+      } 
+      else {
 	 tmp_diskreq = holddiskreq = seg->diskreqlist;
 	 while (tmp_diskreq->seg_next) {
 	    tmp_diskreq = tmp_diskreq->seg_next;
@@ -1154,7 +1225,8 @@ fflush(outputfile);
 	     (currioreq->blkno == seg->endblkno) &&
 	     (seg->size > (currioreq->blkno - (disk_buffer_seg_owner(seg,FALSE))->inblkno))) {
             return_hittype = BUFFER_APPEND;
-	 } else {
+	 } 
+	 else {
             /* check for collision with uncleanable data */
 	    if (disk_buffer_overlap(seg,currioreq) &&
 	        !disk_buffer_reusable_segment_check(currdisk, seg)) {
@@ -1203,51 +1275,51 @@ fflush(outputfile);
 
    if (!seg) {
       fprintf(stderr, "diskreq has NULL segment in disk_buffer_set_segment\n");
-      exit(0);
+      exit(1);
    }
 
    /* check to make sure segment/diskreq combination is valid */
 
    if (is_read && (seg == currdisk->dedicatedwriteseg)) {
       fprintf(stderr, "Read request about to use the write segment\n");
-      exit(0);
+      exit(1);
    }
 
    if (seg->recyclereq && (currdiskreq != seg->recyclereq)) {
       fprintf(stderr, "non-recyclereq detected for recycled segment in disk_buffer_set_segment\n");
-      exit(0);
+      exit(1);
    }
 
    switch (seg->state) {
       case BUFFER_EMPTY:
 	 if (currdiskreq == seg->recyclereq) {
             fprintf(stderr, "seg->state of BUFFER_EMPTY for recyclereq in disk_buffer_set_segment\n");
-            exit(0);
+            exit(1);
 	 }
 	 break;
 
       case BUFFER_CLEAN:
 	 if (currdiskreq == seg->recyclereq) {
             fprintf(stderr, "seg->state of BUFFER_CLEAN for recyclereq in disk_buffer_set_segment\n");
-            exit(0);
+            exit(1);
 	 }
 	 if (is_read && currdiskreq->hittype != BUFFER_NOMATCH && !currdisk->readhitsonwritedata && seg->access && !(seg->access->flags & READ)) {
             fprintf(stderr, "Unallowed read hit on write data disk_buffer_set_segment\n");
-            exit(0);
+            exit(1);
 	 }
 	 break;
 
       case BUFFER_DIRTY:
 	 if (currdiskreq == seg->recyclereq) {
             fprintf(stderr, "seg->state of BUFFER_DIRTY for recyclereq in disk_buffer_set_segment\n");
-            exit(0);
+            exit(1);
 	 }
 	 if (currdiskreq->hittype == BUFFER_NOMATCH) {
             fprintf(stderr, "Attempt to overwrite dirty data in disk_buffer_set_segment\n");
-            exit(0);
+            exit(1);
 	 } else if (is_read && !currdisk->readhitsonwritedata) {
             fprintf(stderr, "Unallowed read hit on write data disk_buffer_set_segment\n");
-            exit(0);
+            exit(1);
 	 }
 	 if (!is_read) {
 	    currdisk->stat.writecombs++;
@@ -1258,12 +1330,12 @@ fflush(outputfile);
 	 if (currdiskreq == seg->recyclereq) {
 	    if (!(currdiskreq->ioreqlist->flags & READ)) {
                fprintf(stderr, "seg->state of BUFFER_READING for write recyclereq in disk_buffer_set_segment\n");
-               exit(0);
+               exit(1);
 	    }
 	 } else if (!is_read || (currdiskreq->hittype == BUFFER_NOMATCH)) {
 	    if (!disk_buffer_stopable_access(currdisk,currdiskreq)) {
                fprintf(stderr, "Attempt to re-target active read segment in disk_buffer_set_segment\n");
-               exit(0);
+               exit(1);
 	    }
          }
 	 break;
@@ -1272,13 +1344,13 @@ fflush(outputfile);
 	 if (currdiskreq == seg->recyclereq) {
 	    if (currdiskreq->ioreqlist->flags & READ) {
                fprintf(stderr, "seg->state of BUFFER_WRITING for read recyclereq in disk_buffer_set_segment\n");
-               exit(0);
+               exit(1);
 	    }
          } else {
 	    if ((currdiskreq->hittype == BUFFER_NOMATCH) || 
 	        (currdiskreq->hittype == BUFFER_PREPEND)) {
                fprintf(stderr, "Attempt to re-target active write segment in disk_buffer_set_segment\n");
-               exit(0);
+               exit(1);
 	    }
 	    if (!is_read) {
 	       currdisk->stat.writecombs++;
@@ -1288,7 +1360,7 @@ fflush(outputfile);
 
       default:
 	 fprintf(stderr, "Invalid segment state in disk_buffer_set_segment - blkno %d, bcount %d, state %d\n", currdiskreq->ioreqlist->blkno, currdiskreq->ioreqlist->bcount, seg->state);
-         exit(0);
+         exit(1);
    }
 
    /* end of sanity checks, now do the work */
@@ -1342,7 +1414,7 @@ fflush(outputfile);
 /*
 	          } else {
                      fprintf(stderr, "Request overlaps with non-clean segment in disk_buffer_set_segment\n");
-                     exit(0);
+                     exit(1);
 */
 	          }
                }
@@ -1364,7 +1436,7 @@ fflush(outputfile);
       seg->diskreqlist = currdiskreq;
       currdiskreq->seg_next = NULL;
    } else {				/* non-empty list */
-      diskreq *prev_diskreq;
+      diskreq *prev_diskreq = 0;
       tmp_diskreq = seg->diskreqlist;
       while (tmp_diskreq) {
 	 if (tmp_diskreq->ioreqlist && 
@@ -1414,7 +1486,7 @@ int disk_buffer_attempt_seg_ownership (disk *currdisk, diskreq *currdiskreq)
    ioreq_event *currioreq = currdiskreq->ioreqlist;
    int         currdiskreq_found = FALSE;
    int         write_found = FALSE;
-   int	       write_incomplete;
+   int	       write_incomplete = 0;
    ioreq_event *tmpioreq;
 
 if (disk_printhack && (simtime >= disk_printhacktime)) {
@@ -1424,12 +1496,12 @@ fflush(outputfile);
 
    if (!seg) {
       fprintf(stderr, "diskreq has NULL segment in disk_buffer_attempt_seg_ownership\n");
-      exit(0);
+      exit(1);
    }
 
    if (currdiskreq->flags & SEG_OWNED) {
       fprintf(stderr, "diskreq already owns seg in disk_buffer_attempt_seg_ownership\n");
-      exit(0);
+      exit(1);
    }
 
    tmp_diskreq = seg->diskreqlist;
@@ -1438,7 +1510,7 @@ fflush(outputfile);
          if (currdiskreq->hittype == BUFFER_PREPEND) {
             if (seg->state != BUFFER_DIRTY) {
                fprintf(stderr, "PREPEND of active request detected in disk_buffer_attempt_seg_ownership\n");
-               exit(0);
+               exit(1);
 	    }
 	    /* release ownership to new request and set up hold areas */
 	    tmp_diskreq->flags &= ~SEG_OWNED;
@@ -1446,7 +1518,7 @@ fflush(outputfile);
 	    if (seg->endblkno != seg->startblkno) {
 	       if (seg->hold_bcount != 0) {
                   fprintf(stderr, "PREPEND of segment already 'holding' data detected in disk_buffer_attempt_seg_ownership\n");
-                  exit(0);
+                  exit(1);
 	       }
                seg->hold_blkno = seg->startblkno;
 	       seg->hold_bcount = seg->endblkno - seg->startblkno;
@@ -1514,8 +1586,9 @@ fflush(outputfile);
 	    while (currioreq->next) {
 	       currioreq = currioreq->next;
 	    }
-	    seg->minreadaheadblkno = max(seg->minreadaheadblkno, min((currioreq->blkno + currioreq->bcount + currdisk->minreadahead), currdisk->numblocks));
-	    seg->maxreadaheadblkno = max(seg->maxreadaheadblkno, min(disk_buffer_get_max_readahead(currdisk,seg,currioreq), currdisk->numblocks));
+	    seg->minreadaheadblkno = max(seg->minreadaheadblkno, min((currioreq->blkno + currioreq->bcount + currdisk->minreadahead), currdisk->model->dm_sectors));
+
+	    seg->maxreadaheadblkno = max(seg->maxreadaheadblkno, min(disk_buffer_get_max_readahead(currdisk,seg,currioreq), currdisk->model->dm_sectors));
 
 	    if ((seg->endblkno >= (currioreq->blkno + currioreq->bcount)) &&
 	        (currdisk->effectivehda == currdiskreq)) {
@@ -1568,7 +1641,7 @@ int disk_buffer_get_max_readahead (disk *currdisk, segment *seg, ioreq_event *cu
 
    if (!(curr->flags & READ)) {
       fprintf(stderr, "No read-ahead for write accesses, in disk_buffer_get_max_readahead\n");
-      exit(0);
+      exit(1);
    }
 
    endreq = curr->blkno + curr->bcount;
@@ -1578,23 +1651,31 @@ int disk_buffer_get_max_readahead (disk *currdisk, segment *seg, ioreq_event *cu
    } else if (currdisk->contread == BUFFER_DEC_PREFETCH_SCHEME) {
       int startlbn;
       int endlbn;
-      int cylno;
-      int surface;
-      int blkno;
-      band* bandptr;
+      struct dm_pbn pbn;
+      //      band* bandptr;
 
       curr->blkno += curr->bcount;
 
-      bandptr = disk_translate_lbn_to_pbn(currdisk, curr->blkno, MAP_FULL, 
-					  &cylno, &surface, &blkno);
+      // bandptr = 
+      // was MAP_FULL
+      currdisk->model->layout->dm_translate_ltop(currdisk->model, 
+						 curr->blkno, 
+						 MAP_FULL,
+						 &pbn, 
+						 0); // remapsector
 
-      disk_get_lbn_boundaries_for_track(currdisk, bandptr, cylno, surface, 
-					&startlbn, &endlbn);
+      currdisk->model->layout->dm_get_track_boundaries(currdisk->model,
+						       &pbn,
+						       &startlbn, 
+						       &endlbn,
+						       0); // remapsector
+      // new track_boundaries semantics
+      endlbn++;
 
       curr->blkno -= curr->bcount;
 
       if (seg->startblkno < startlbn) {
-          return(endlbn + bandptr->blkspertrack);
+	return (endlbn + currdisk->model->layout->dm_get_sectors_lbn(currdisk->model, curr->blkno));
       } else {
 	  return(endlbn);
       }
@@ -1604,9 +1685,11 @@ int disk_buffer_get_max_readahead (disk *currdisk, segment *seg, ioreq_event *cu
       return(endreq + currdisk->maxreadahead);
    }
    if (currdisk->keeprequestdata > 0) {
-      if ((curr->bcount >= seg->size) || ((seg->size - curr->bcount) < currdisk->minreadahead)) {
-	 return(endreq + currdisk->minreadahead);
-      }
+      if ((curr->bcount >= seg->size) 
+	  || ((seg->size - curr->bcount) < currdisk->minreadahead)) 
+	{
+	  return(endreq + currdisk->minreadahead);
+	}
       return(curr->blkno + seg->size);
    }
 
