@@ -1,3 +1,35 @@
+/*
+ * DiskSim Storage Subsystem Simulation Environment (Version 4.0)
+ * Revision Authors: John Bucy, Greg Ganger
+ * Contributors: John Griffin, Jiri Schindler, Steve Schlosser
+ *
+ * Copyright (c) of Carnegie Mellon University, 2001-2008.
+ *
+ * This software is being provided by the copyright holders under the
+ * following license. By obtaining, using and/or copying this software,
+ * you agree that you have read, understood, and will comply with the
+ * following terms and conditions:
+ *
+ * Permission to reproduce, use, and prepare derivative works of this
+ * software is granted provided the copyright and "No Warranty" statements
+ * are included with all reproductions and derivative works and associated
+ * documentation. This software may also be redistributed without charge
+ * provided that the copyright and "No Warranty" statements are included
+ * in all redistributions.
+ *
+ * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
+ * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
+ * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
+ * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
+ * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
+ * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
+ * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
+ * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
+ * OR DOCUMENTATION.
+ *
+ */
+
+
 
 /*
  * DiskSim Storage Subsystem Simulation Environment (Version 2.0)
@@ -72,6 +104,15 @@
 #include "disksim_pfsim.h"
 #include "disksim_ioface.h"
 #include "disksim_synthio.h"
+#include "config.h"
+
+#include "modules/modules.h"
+
+
+int pf_how_many_cpus()
+{
+   return (numcpus);
+}
 
 
 static void pf_allocate_process_structs()
@@ -222,27 +263,38 @@ static void pf_add_to_pendiolist (process *procp, ioreq_event *curr)
 }
 
 
-event * pf_io_done_notify (ioreq_event *curr)
+event * pf_io_done_notify (ioreq_event *curr, void *ctx)
 {
    ioreq_event *tmp = pendiolist;
    ioreq_event *tmp2;
    wakeup_event *tmpwake = NULL;
 
-if (pf_printhack) {
-   fprintf (outputfile, "pf_io_done_notify: curr->buf %p, curr->opid %x, curr->blkno %d\n", curr->buf, curr->opid, curr->blkno);
-}
+   if(disksim->external_control){
+     disksim->external_io_done_notify (curr, ctx);
+   }
+
+   if (pf_printhack) {
+     fprintf (outputfile, "pf_io_done_notify: curr->buf %p, curr->opid %x, curr->blkno %d\n", curr->buf, curr->opid, curr->blkno);
+   }
+
    ASSERT(pendiolist != NULL);
    if ((tmp->buf == curr->buf) && (tmp->opid == curr->opid)) {
       pendiolist = tmp->next;
-   } else {
-      while ((tmp->next) && ((tmp->next->buf != curr->buf) || (tmp->next->opid != curr->opid))) {
+   } 
+   else {
+     while ((tmp->next) && 
+	    ((tmp->next->buf != curr->buf) || 
+	     (tmp->next->opid != curr->opid))) 
+       {
          tmp = tmp->next;
-      }
-      ASSERT(tmp->next != NULL);
-      tmp2 = tmp->next;
-      tmp->next = tmp2->next;
-      tmp = tmp2;
+       }
+     
+     ASSERT(tmp->next != NULL);
+     tmp2 = tmp->next;
+     tmp->next = tmp2->next;
+     tmp = tmp2;
    }
+
    if (tmp->flags & PF_BUF_WANTED) {
       tmpwake = (wakeup_event *) getfromextraq();
       tmpwake->type = WAKEUP_EVENT;
@@ -251,16 +303,21 @@ if (pf_printhack) {
       tmpwake->chan = tmp->buf;
       tmpwake->dropit = 0;
    }
+
    if (curr->flags & TIME_CRITICAL) {
       stat_update(&timecritrespstats, (simtime - tmp->time));
-   } else if (curr->flags & TIME_LIMITED) {
+   } 
+   else if (curr->flags & TIME_LIMITED) {
       stat_update(&timelimitrespstats, (simtime - tmp->time));
-   } else {
+   } 
+   else {
       stat_update(&timenoncritrespstats, (simtime - tmp->time));
    }
+
    if (!(curr->flags & TIME_LIMITED) || (tmpwake)) {
       addtoextraq((event *) tmp);
-   } else {
+   } 
+   else {
       tmp->next = doneiolist;
       doneiolist = tmp;
    }
@@ -277,7 +334,7 @@ static int pf_iowait (void *chan, process *procp)
       tmp = tmp->next;
    }
    if (tmp == NULL) {
-      ioreq_event *prev;
+      ioreq_event *prev = 0;
       tmp = doneiolist;
       while ((tmp) && (tmp->buf != chan)) {
          prev = tmp;
@@ -315,7 +372,7 @@ fprintf (outputfile, "pf_iowait: chan %p, read %d, crit %d, opid %d, blkno %x\n"
    tmptime = io_raise_priority(tmp->opid, tmp->devno, tmp->blkno, tmp->buf);
    if (tmptime != 0.0) {
       fprintf(stderr, "Haven't handled non-zero bufferwait time yet\n");
-      exit(0);
+      exit(1);
    }
    return(TRUE);
 }
@@ -374,20 +431,22 @@ static double pf_get_time_to_next_intr_event (intr_event *intrp)
 
 void pf_idle_cpu_recheck_dispq (int cpunum)
 {
-   ASSERT(cpus[cpunum].current->procp == NULL);
+  ASSERT(cpus[cpunum].current->procp == NULL);
+
    if (cpus[cpunum].idleevents == NULL) {
       event *new = getfromextraq();
       new->type = IDLELOOP_EVENT;
       new->time = PF_IDLE_CHECK_DISPQ_TIME;
       new->next = cpus[cpunum].idleevents;
       cpus[cpunum].idleevents = new;
+
       if (cpus[cpunum].current->intrp == NULL) {
-         cpus[cpunum].idletime += simtime - cpus[cpunum].idlestart;
-         cpus[cpunum].falseidletime += pf_add_false_idle_time(simtime - cpus[cpunum].idlestart);
-         cpus[cpunum].current->time = simtime + (cpus[cpunum].scale * new->time);
-         new->time = simtime;
-         cpus[cpunum].state = CPU_IDLE_WORK;
-         addtointq((event *) cpus[cpunum].current);
+	cpus[cpunum].idletime += simtime - cpus[cpunum].idlestart;
+	cpus[cpunum].falseidletime += pf_add_false_idle_time(simtime - cpus[cpunum].idlestart);
+	cpus[cpunum].current->time = simtime + (cpus[cpunum].scale * new->time);
+	new->time = simtime;
+	cpus[cpunum].state = CPU_IDLE_WORK;
+	addtointq((event *) cpus[cpunum].current);
       }
    }
 }
@@ -409,10 +468,13 @@ static void pf_handle_cswitch_event (cswitch_event *curr, cpu_event *cpu_ev)
 
    /* Used to identify how long the system has been completely idle   */
    /* Simulation should stop rather than continuing to idle endlessly */
-   if ((idlereset == 0) && ((curr->newprocp == NULL) || (curr->newprocp->idler != 0))) {
+   if ((idlereset == 0) && 
+       ((curr->newprocp == NULL) || 
+	(curr->newprocp->idler != 0))) {
       idlein = simtime;
       idlereset = 1;
-   } else {
+   } 
+   else {
       idlereset = 0;
    }
 
@@ -474,29 +536,35 @@ static void pf_handle_sleep_event (sleep_event *curr, cpu_event *cpu_ev)
 }
 
 
-static void pf_handle_wakeup_event (wakeup_event *curr, cpu_event *cpu_ev)
+static void pf_handle_wakeup_event(wakeup_event *curr, cpu_event *cpu_ev)
 {
    process *procp;
 
-   if (pf_printhack)
-   fprintf (outputfile, "%f\tWAKEUP cpu=%d chan=%p info=%d\n", simtime, cpu_ev->cpunum, curr->chan, curr->info);
+   if (pf_printhack) {
+     fprintf (outputfile, "%f\tWAKEUP cpu=%d chan=%p info=%d\n", 
+	      simtime, cpu_ev->cpunum, curr->chan, curr->info);
+   }
 
    while ((procp = pf_disp_get_from_sleep_queue(curr->chan))) {
 
-if (pf_printhack) {
-fprintf (outputfile, "Waking up process %d\n", procp->pid);
-}
-      procp->runsleep += simtime - procp->lastsleep;
-      if (procp->iosleep == 1) {
-         procp->runiosleep += simtime - procp->lastsleep;
-      }
-      pf_disp_wakeup(procp);
+     if (pf_printhack) {
+       fprintf (outputfile, "Waking up process %d\n", procp->pid);
+     }
+
+     procp->runsleep += simtime - procp->lastsleep;
+     if (procp->iosleep == 1) {
+       procp->runiosleep += simtime - procp->lastsleep;
+     }
+     pf_disp_wakeup(procp);
    }
 
    procp = cpu_ev->procp;
-   if ((procp) && ((procp->pfflags & PF_SLEEP_FLAG) && (procp->chan == curr->chan))) {
-      procp->pfflags |= PF_WAKED;
-   }
+   if ((procp) 
+       && ((procp->pfflags & PF_SLEEP_FLAG) 
+	   && (procp->chan == curr->chan))) 
+     {
+       procp->pfflags |= PF_WAKED;
+     }
 }
 
 
@@ -598,7 +666,7 @@ static void pf_handle_intend_event (intend_event *curr, cpu_event *cpu_ev)
       pf_clock_interrupt_complete(intrp);
    } else {
       fprintf (stderr, "pf_handle_intend_event: unknown interrupt vector (%d)\n", intrp->vector);
-      exit (0);
+      exit(1);
    }
    intrp = cpu_ev->intrp;
    if (intrp == NULL) {
@@ -631,48 +699,49 @@ fprintf (outputfile, "%f, SYNTHIO_EVENT\n", simtime);
 
 static void pf_handle_event (event *curr, cpu_event *cpu_ev)
 {
-if (pf_printhack) {
-fprintf (outputfile, "%f, entered pf_handle_event with event type %d\n", simtime, curr->type);
-fflush(outputfile);
-}
-   switch (curr->type) {
-      case IDLELOOP_EVENT:
-			pf_handle_idleloop_event((idleloop_event *)curr, cpu_ev);
-			break;
-      case CSWITCH_EVENT:
-			pf_handle_cswitch_event((cswitch_event *)curr, cpu_ev);
-			break;
-      case SLEEP_EVENT:
-			pf_handle_sleep_event((sleep_event *)curr, cpu_ev);
-			break;
-      case WAKEUP_EVENT:
-			pf_handle_wakeup_event((wakeup_event *)curr, cpu_ev);
-			break;
-      case IOREQ_EVENT:
-			pf_handle_ioreq_event((ioreq_event *)curr, cpu_ev);
-			curr = NULL;
-			break;
-      case IOACC_EVENT:
-			pf_handle_ioacc_event((ioreq_event *)curr, cpu_ev);
-			curr = NULL;
-			break;
-      case INTEND_EVENT:
-			pf_handle_intend_event((intend_event *)curr, cpu_ev);
-			break;
-      case SYNTHIO_EVENT:
-			pf_handle_synthio_event(curr, cpu_ev);
-			curr = NULL;
-			break;
-      default:
-			if ((curr->type >= IO_MIN_EVENT) && (curr->type <= IO_MAX_EVENT)) {
-			   pf_handle_io_internal_event((ioreq_event *)curr, cpu_ev);
-			   curr = NULL;
-			   break;
-			}
-			fprintf(stderr, "Trying to handle unknown pf event - %d\n", curr->type);
-			exit(0);
-   }
-   addtoextraq((event *) curr);
+  if (pf_printhack) {
+    fprintf (outputfile, "%f, entered pf_handle_event with event type %d\n", simtime, curr->type);
+    fflush(outputfile);
+  }
+
+  switch (curr->type) {
+  case IDLELOOP_EVENT:
+    pf_handle_idleloop_event((idleloop_event *)curr, cpu_ev);
+    break;
+  case CSWITCH_EVENT:
+    pf_handle_cswitch_event((cswitch_event *)curr, cpu_ev);
+    break;
+  case SLEEP_EVENT:
+    pf_handle_sleep_event((sleep_event *)curr, cpu_ev);
+    break;
+  case WAKEUP_EVENT:
+    pf_handle_wakeup_event((wakeup_event *)curr, cpu_ev);
+    break;
+  case IOREQ_EVENT:
+    pf_handle_ioreq_event((ioreq_event *)curr, cpu_ev);
+    curr = NULL;
+    break;
+  case IOACC_EVENT:
+    pf_handle_ioacc_event((ioreq_event *)curr, cpu_ev);
+    curr = NULL;
+    break;
+  case INTEND_EVENT:
+    pf_handle_intend_event((intend_event *)curr, cpu_ev);
+    break;
+  case SYNTHIO_EVENT:
+    pf_handle_synthio_event(curr, cpu_ev);
+    curr = NULL;
+    break;
+  default:
+    if ((curr->type >= IO_MIN_EVENT) && (curr->type <= IO_MAX_EVENT)) {
+      pf_handle_io_internal_event((ioreq_event *)curr, cpu_ev);
+      curr = NULL;
+      break;
+    }
+    fprintf(stderr, "Trying to handle unknown pf event - %d\n", curr->type);
+    exit(1);
+  }
+  addtoextraq((event *) curr);
 }
 
 
@@ -735,7 +804,7 @@ void pf_handle_intr_event (intr_event *intrp, int cpunum)
    } else {
       if (removefromintq((event *)cpu_ev) == 0) {
          fprintf(stderr, "Event from active CPU not pending on internal queue\n");
-         exit(0);
+         exit(1);
       }
       if (intrp->next) {
          intrp->next->runtime += simtime - intrp->next->eventlist->time;
@@ -757,7 +826,7 @@ void pf_handle_intr_event (intr_event *intrp, int cpunum)
       pf_handle_clock_intr_arrive(intrp, cpu_ev);
    } else {
       fprintf(stderr, "Unknown interrupt vector: %d\n", intrp->vector);
-      exit(0);
+      exit(1);
    }
    cpu_ev->time = currcpu->scale * pf_get_time_to_next_intr_event(intrp);
    cpu_ev->time += simtime;
@@ -773,53 +842,62 @@ static void pf_handle_cpu_event (cpu_event *curr)
    event *tmp;
    cpu *currcpu = &cpus[(curr->cpunum)];
 
-/*
-if (simtime > 29000.0) {
-   pf_printhack = 1;
-}
-*/
-   if (pf_printhack) {
-   fprintf (outputfile, "Entered handle_cpu_event: %d\n", curr->type);
-   fflush(outputfile);
+   /*
+    * if (simtime > 29000.0) {
+    *   pf_printhack = 1;
+    * }
+    */
+
+   if(pf_printhack) {
+     fprintf (outputfile, "Entered handle_cpu_event: %d\n", curr->type);
+     fflush(outputfile);
    }
 
-   if (curr->intrp) {
+   if(curr->intrp) {
       tmp = curr->intrp->eventlist;
       curr->intrp->runtime += simtime - tmp->time;
       curr->intrp->eventlist = tmp->next;
-   } else if (curr->procp) {
+   } 
+   else if(curr->procp) {
       curr->procp->lasteventtime = simtime;
       tmp = curr->procp->eventlist;
       curr->procp->runtime += simtime - tmp->time;
       curr->procp->eventlist = tmp->next;
-   } else {
+   } 
+   else {
       tmp = currcpu->idleevents;
       currcpu->idleworktime += simtime - tmp->time;
       currcpu->idleevents = tmp->next;
    }
+
    ASSERT(tmp != NULL);
    tmp->next = NULL;
    pf_handle_event(tmp, curr);
-   if (curr->intrp) {
+
+   if(curr->intrp) {
       curr->time = pf_get_time_to_next_intr_event(curr->intrp);
-   } else if (curr->procp == NULL) {
-      if (currcpu->idleevents == NULL) {
-         currcpu->state = CPU_IDLE;
-         currcpu->idlestart = simtime;
-         return;
-      }
-      curr->time = currcpu->idleevents->time;
-      currcpu->idleevents->time = simtime;
-      currcpu->state = CPU_IDLE_WORK;
-   } else {
-      curr->time = pf_get_time_to_next_process_event(curr->procp);
+   } 
+   else if(curr->procp == NULL) {
+     if (currcpu->idleevents == NULL) {
+       currcpu->state = CPU_IDLE;
+       currcpu->idlestart = simtime;
+       return;
+     }
+     curr->time = currcpu->idleevents->time;
+     currcpu->idleevents->time = simtime;
+     currcpu->state = CPU_IDLE_WORK;
+   } 
+   else {
+     curr->time = pf_get_time_to_next_process_event(curr->procp);
    }
+
    curr->time = simtime + (currcpu->scale * curr->time);
    addtointq((event *) curr);
 
-   if (pf_printhack) {
-   fprintf (outputfile, "Exited handle_cpu_event\n");
-   fflush(outputfile);
+   
+   if(pf_printhack) {
+     fprintf(outputfile, "Exited handle_cpu_event\n");
+     fflush(outputfile);
    }
 }
 
@@ -892,12 +970,12 @@ void pf_initialize (int seedvalue)
 /*
 fprintf (outputfile, "Entered pf_initialize - numcpus %d, pfscale %f, pfinit %d\n", numcpus, pfscale, (pfinitfile != NULL));
 */
-    StaticAssert (sizeof(cpu_event) <= DISKSIM_EVENT_SIZE);
-    StaticAssert (sizeof(intend_event) <= DISKSIM_EVENT_SIZE);
-    StaticAssert (sizeof(cswitch_event) <= DISKSIM_EVENT_SIZE);
-    StaticAssert (sizeof(idleloop_event) <= DISKSIM_EVENT_SIZE);
-    StaticAssert (sizeof(sleep_event) <= DISKSIM_EVENT_SIZE);
-    StaticAssert (sizeof(wakeup_event) <= DISKSIM_EVENT_SIZE);
+   StaticAssert (sizeof(cpu_event) <= DISKSIM_EVENT_SIZE);
+   StaticAssert (sizeof(intend_event) <= DISKSIM_EVENT_SIZE);
+   StaticAssert (sizeof(cswitch_event) <= DISKSIM_EVENT_SIZE);
+   StaticAssert (sizeof(idleloop_event) <= DISKSIM_EVENT_SIZE);
+   StaticAssert (sizeof(sleep_event) <= DISKSIM_EVENT_SIZE);
+   StaticAssert (sizeof(wakeup_event) <= DISKSIM_EVENT_SIZE);
 
    stat_initialize(statdeffile, "Response time", &timecritrespstats);
    stat_initialize(statdeffile, "Response time", &timelimitrespstats);
@@ -913,81 +991,77 @@ fprintf (outputfile, "Entered pf_initialize - numcpus %d, pfscale %f, pfinit %d\
 
    DISKSIM_srand48(seedvalue);
    procp = synthlist;
-   while (procp) {
-      if (procp->space) {
-         synthio_initialize_generator(procp);
-      }
-      procp = procp->next;
+   while(procp) {
+     if(procp->space) {
+       synthio_initialize_generator(procp);
+     }
+     procp = procp->next;
    }
+ 
+   // initial distribution of processes to cpus
    pf_dispatcher_init(synthlist);
    synthlist = NULL;
-   for (i=0; i<numcpus; i++) {
-fprintf (outputfile, "Kicking off cpu #%d\n", i);
-      if (cpus[i].current->procp) {
-	 cpus[i].current->procp->active = 1;
-         cpus[i].current->time = cpus[i].scale * pf_get_time_to_next_process_event(cpus[i].current->procp);
-         addtointq((event *) cpus[i].current);
+
+   for (i = 0; i < numcpus; i++) {
+     fprintf (outputfile, "Kicking off cpu #%d\n", i);
+
+     if(cpus[i].current->procp) {
+       cpus[i].current->procp->active = 1;
+       cpus[i].current->time = cpus[i].scale * 
+	 pf_get_time_to_next_process_event(cpus[i].current->procp);
+
+       addtointq((event *)cpus[i].current);
          cpus[i].state = CPU_PROCESS;
-fprintf (outputfile, "First event occurs at time %f\n", cpus[i].current->time);
-      } else {
-         cpus[i].state = CPU_IDLE;
-      }
+	 fprintf (outputfile, "First event occurs at time %f\n", 
+		  cpus[i].current->time);
+     } 
+     else {
+       cpus[i].state = CPU_IDLE;
+     }
    }
 }
 
 
-void pf_param_override (int structtype, char *paramname, char *paramval, int first, int last)
-{
-   switch (structtype) {
 
-      case SYNTHIO_OVERRIDE:
-                       synthio_param_override(paramname, paramval, first, last);
-		       break;
+int disksim_pf_loadparams(struct lp_block *b) {
+  pf_info_t *pf_info;
 
-      default:         fprintf(stderr, "Unsupported structure type at pf_param_override: %d\n", structtype);
-		       exit(0);
-   }
+  if(!disksim->pf_info) {
+    pf_info = DISKSIM_malloc (sizeof(pf_info_t));
+    if(!pf_info) return 0;
+    bzero ((char *)pf_info, sizeof(pf_info_t));
+    disksim->pf_info = pf_info;
+  }
+
+/*    unparse_block(b, outputfile); */
+
+  //#include "modules/disksim_pf_param.c"
+  lp_loadparams(0, b, &disksim_pf_mod);
+
+  return 1;
 }
 
 
-void pf_readparams (FILE *parfile)
-{
-   pf_info_t *pf_info = DISKSIM_malloc (sizeof(pf_info_t));
-   bzero ((char *)pf_info, sizeof(pf_info_t));
-   disksim->pf_info = pf_info;
+int disksim_pf_stats_loadparams(struct lp_block *b) {
+  pf_info_t *pf_info;
 
-   /* some important initialization (these are remapped into the new pf_info) */
-   curlbolt = 1;
-   lastuser = 0.0;
-   idlein = 0.0;
+  if(!disksim->pf_info) {
+    pf_info = DISKSIM_malloc (sizeof(pf_info_t));
+    if(!pf_info) return 0;
+    bzero ((char *)pf_info, sizeof(pf_info_t));
+    disksim->pf_info = pf_info;
+  }
 
-   fscanf(parfile, "\nProcess-Flow Input Parameters\n");
-   fprintf (outputfile, "\nProcess-Flow Input Parameters\n");
-   fscanf(parfile, "-----------------------------\n");
-   fprintf (outputfile, "-----------------------------\n");
 
-   fscanf(parfile, "\nPRINTED PROCESS-FLOW STATISTICS\n\n");
-   fprintf (outputfile, "\nPRINTED PROCESS-FLOW STATISTICS\n\n");
+  //#include "modules/disksim_pf_stats_param.c"
+  lp_loadparams(0, b, &disksim_pf_stats_mod);
 
-   getparam_bool(parfile, "Print per-process stats?",&pf_print_perprocessstats);
-   getparam_bool(parfile, "Print per-CPU stats?", &pf_print_percpustats);
-   getparam_bool(parfile, "Print all interrupt stats?", &pf_print_intrstats);
-   getparam_bool(parfile, "Print sleep stats?", &pf_print_sleepstats);
 
-   fscanf(parfile, "\nGENERAL PROCESS-FLOW PARAMETERS\n");
-   fprintf (outputfile, "\nGENERAL PROCESS-FLOW PARAMETERS\n");
-
-   getparam_int(parfile, "Number of processors", &numcpus, 1, 1, 0);
-   cpus = (cpu *) DISKSIM_malloc (numcpus * sizeof(cpu));
-   getparam_double(parfile, "Process-Flow Time Scale", &pfscale, 1, (double) 0.0000000001, (double) 0.0);
-
-   fscanf(parfile, "\nSYNTHETIC I/O TRACE PARAMETERS\n\n");
-   fprintf (outputfile, "\nSYNTHETIC I/O TRACE PARAMETERS\n\n");
-
-   synthio_readparams(parfile);
+  return 1;
 }
 
 
+
 void pf_cleanstats()
 {
    int i;
@@ -1016,7 +1090,7 @@ void pf_cleanstats()
                       break;
          default:
                       fprintf(stderr, "Unknown CPU state at pf_cleanstats - %d\n", cpus[i].state);
-                      exit(0);
+                      exit(1);
       }
    }
    procp = process_livelist;
@@ -1179,9 +1253,9 @@ static void pf_print_process_stats()
       fprintf(outputfile, "Process %d Number of read requests: %d\n", procp->pid, procp->ioreads);
       fprintf(outputfile, "Process %d Number of C-switches:    %d\n", procp->pid, procp->cswitches);
       fprintf(outputfile, "Process %d Number of sleeps:        %d\n", procp->pid, procp->sleeps);
-      fprintf(outputfile, "Process %d Average sleep time:      %f\n", procp->pid, (procp->runsleep / (double) procp->sleeps));
+      fprintf(outputfile, "Process %d Average sleep time:      %f\n", procp->pid, (procp->runsleep / (double) max(procp->sleeps,1)));
       fprintf(outputfile, "Process %d Number of I/O sleeps:    %d\n", procp->pid, procp->iosleeps);
-      fprintf(outputfile, "Process %d Average I/O sleep time:  %f\n", procp->pid, (procp->runiosleep / (double) procp->iosleeps));
+      fprintf(outputfile, "Process %d Average I/O sleep time:  %f\n", procp->pid, (procp->runiosleep / (double) max(procp->iosleeps,1)));
       fprintf(outputfile, "Process %d False idle time:         %f\n", procp->pid, procp->falseidletime);
       fprintf(outputfile, "Process %d Read Time limits measured: %d\n", procp->pid, stat_get_count(&procp->readtimelimitstats));
       sprintf(procstr, "Process %d Read ", procp->pid);
@@ -1279,8 +1353,8 @@ void pf_printstats()
    fprintf (outputfile, "\nPROCESS FLOW STATISTICS\n");
    fprintf (outputfile, "-----------------------\n\n");
 
-   pf_print_pf_stats(0, (numcpus-1));
-   pf_print_interrupt_stats(0, (numcpus-1));
+   pf_print_pf_stats(0, (numcpus - 1));
+   pf_print_interrupt_stats(0, (numcpus - 1));
    pf_print_sleep_stats();
    pf_print_percpu_stats();
    pf_print_process_stats();
