@@ -1,3 +1,35 @@
+/*
+ * DiskSim Storage Subsystem Simulation Environment (Version 4.0)
+ * Revision Authors: John Bucy, Greg Ganger
+ * Contributors: John Griffin, Jiri Schindler, Steve Schlosser
+ *
+ * Copyright (c) of Carnegie Mellon University, 2001-2008.
+ *
+ * This software is being provided by the copyright holders under the
+ * following license. By obtaining, using and/or copying this software,
+ * you agree that you have read, understood, and will comply with the
+ * following terms and conditions:
+ *
+ * Permission to reproduce, use, and prepare derivative works of this
+ * software is granted provided the copyright and "No Warranty" statements
+ * are included with all reproductions and derivative works and associated
+ * documentation. This software may also be redistributed without charge
+ * provided that the copyright and "No Warranty" statements are included
+ * in all redistributions.
+ *
+ * NO WARRANTY. THIS SOFTWARE IS FURNISHED ON AN "AS IS" BASIS.
+ * CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER
+ * EXPRESSED OR IMPLIED AS TO THE MATTER INCLUDING, BUT NOT LIMITED
+ * TO: WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY
+ * OF RESULTS OR RESULTS OBTAINED FROM USE OF THIS SOFTWARE. CARNEGIE
+ * MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT
+ * TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.
+ * COPYRIGHT HOLDERS WILL BEAR NO LIABILITY FOR ANY USE OF THIS SOFTWARE
+ * OR DOCUMENTATION.
+ *
+ */
+
+
 
 /*
  * DiskSim Storage Subsystem Simulation Environment (Version 2.0)
@@ -66,57 +98,17 @@
  * holders.
  */
 
-#include "disksim_global.h"
-#include "disksim_iosim.h"
-#include "disksim_stat.h"
+
 #include "disksim_simpledisk.h"
-#include "disksim_ioqueue.h"
-#include "disksim_bus.h"
+
+
+#include "modules/disksim_simpledisk_param.h"
 
 
 /* read-only globals used during readparams phase */
 static char *statdesc_acctimestats	=	"Access time";
 
 
-typedef struct {
-   statgen acctimestats;
-   double  requestedbus;
-   double  waitingforbus;
-   int     numbuswaits;
-} simpledisk_stat_t;
-
-
-typedef struct simpledisk {
-   double acctime;
-   double overhead;
-   double bus_transaction_latency;
-   int numblocks;
-   int devno;
-   int inited;
-   struct ioq *queue;
-   int media_busy;
-   int reconnect_reason;
-
-   double blktranstime;
-   int maxqlen;
-   int busowned;
-   ioreq_event *buswait;
-   int neverdisconnect;
-   int numinbuses;
-   int inbuses[MAXINBUSES];
-   int depth[MAXINBUSES];
-   int slotno[MAXINBUSES];
-
-   int printstats;
-   simpledisk_stat_t stat;
-} simpledisk_t;
-
-
-typedef struct simpledisk_info {
-   struct simpledisk *simpledisks;
-   int numsimpledisks;
-} simplediskinfo_t;
-
 
 /* private remapping #defines for variables from device_info_t */
 #define numsimpledisks         (disksim->simplediskinfo->numsimpledisks)
@@ -124,10 +116,10 @@ typedef struct simpledisk_info {
 
 
 
-static struct simpledisk *getsimpledisk (int devno)
+struct simpledisk *getsimpledisk (int devno)
 {
    ASSERT1((devno >= 0) && (devno < MAXDEVICES), "devno", devno);
-   return (&disksim->simplediskinfo->simpledisks[devno]);
+   return (disksim->simplediskinfo->simpledisks[devno]);
 }
 
 
@@ -137,11 +129,12 @@ int simpledisk_set_depth (int devno, int inbusno, int depth, int slotno)
    int cnt;
 
    currdisk = getsimpledisk (devno);
+   assert(currdisk);
    cnt = currdisk->numinbuses;
    currdisk->numinbuses++;
    if ((cnt + 1) > MAXINBUSES) {
       fprintf(stderr, "Too many inbuses specified for simpledisk %d - %d\n", devno, (cnt+1));
-      exit(0);
+      exit(1);
    }
    currdisk->inbuses[cnt] = inbusno;
    currdisk->depth[cnt] = depth;
@@ -210,16 +203,15 @@ int simpledisk_get_busno (ioreq_event *curr)
 
 
 /*
-  **- simpledisk_send_event_up_path
-
-  Acquires the bus (if not already acquired), then uses bus_delay to send the event
-    up the path.
-
-  If the bus is already owned by this device or can be acquired immediately (interleaved
-    bus), the event is sent immediately.  Otherwise, disk_bus_ownership_grant will later
-    send the event.
-	
-  */
+ * simpledisk_send_event_up_path()
+ *
+ * Acquires the bus (if not already acquired), then uses bus_delay to
+ * send the event up the path.
+ *
+ * If the bus is already owned by this device or can be acquired
+ * immediately (interleaved bus), the event is sent immediately.
+ * Otherwise, disk_bus_ownership_grant will later send the event.
+ */
   
 static void simpledisk_send_event_up_path (ioreq_event *curr, double delay)
 {
@@ -252,16 +244,16 @@ static void simpledisk_send_event_up_path (ioreq_event *curr, double delay)
          /* Remember when we started waiting (only place this is written) */
 	 currdisk->stat.requestedbus = simtime;
       } else {
-         bus_delay(busno, DISK, curr->devno, delay, curr); /* Never for SCSI */
+         bus_delay(busno, DEVICE, curr->devno, delay, curr); /* Never for SCSI */
       }
    } else if (currdisk->busowned == busno) {
 
       //fprintf (outputfile, "Already own bus - so send it on up\n");
 
-      bus_delay(busno, DISK, curr->devno, delay, curr);
+      bus_delay(busno, DEVICE, curr->devno, delay, curr);
    } else {
       fprintf(stderr, "Wrong bus owned for transfer desired\n");
-      exit(0);
+      exit(1);
    }
 }
 
@@ -287,13 +279,13 @@ void simpledisk_bus_ownership_grant (int devno, ioreq_event *curr, int busno, do
    }
    if (tmp == NULL) {
       fprintf(stderr, "Bus ownership granted to unknown simpledisk request - devno %d, busno %d\n", devno, busno);
-      exit(0);
+      exit(1);
    }
    currdisk->busowned = busno;
    currdisk->stat.waitingforbus += arbdelay;
    //ASSERT (arbdelay == (simtime - currdisk->stat.requestedbus));
    currdisk->stat.numbuswaits++;
-   bus_delay(busno, DISK, devno, tmp->time, tmp);
+   bus_delay(busno, DEVICE, devno, tmp->time, tmp);
 }
 
 
@@ -317,7 +309,7 @@ void simpledisk_bus_delay_complete (int devno, ioreq_event *curr, int sentbusno)
       }
       if (tmp->next != curr) {
          fprintf(stderr, "Bus delay complete for unknown simpledisk request - devno %d, busno %d\n", devno, busno.value);
-         exit(0);
+         exit(1);
       }
       tmp->next = curr->next;
    }
@@ -346,7 +338,7 @@ static void simpledisk_request_complete(ioreq_event *curr)
 
    if ((curr = ioqueue_physical_access_done(currdisk->queue,curr)) == NULL) {
       fprintf(stderr, "simpledisk_request_complete:  ioreq_event not found by ioqueue_physical_access_done call\n");
-      exit(0);
+      exit(1);
    }
 
    /* send completion interrupt */
@@ -437,7 +429,7 @@ static void simpledisk_request_arrive (ioreq_event *curr)
    if ((curr->blkno < 0) || (curr->bcount <= 0) ||
        ((curr->blkno + curr->bcount) > currdisk->numblocks)) {
       fprintf(stderr, "Invalid set of blocks requested from simpledisk - blkno %d, bcount %d, numblocks %d\n", curr->blkno, curr->bcount, currdisk->numblocks);
-      exit(0);
+      exit(1);
    }
 
    /* create a new request, set it up for initial interrupt */
@@ -582,8 +574,7 @@ static void simpledisk_interrupt_complete (ioreq_event *curr)
 	 break;
 
       default:
-         fprintf(stderr, "Unrecognized event type at simpledisk_interrupt_complete\n");
-         exit(0);
+         ddbg_assert2(0, "bad event type");
    }
 }
 
@@ -630,7 +621,7 @@ void simpledisk_event_arrive (ioreq_event *curr)
 
       default:
          fprintf(stderr, "Unrecognized event type at simpledisk_event_arrive\n");
-         exit(0);
+         exit(1);
    }
 
    // fprintf (outputfile, "Exiting simpledisk_event_arrive\n");
@@ -657,7 +648,7 @@ void simpledisk_get_mapping (int maptype, int devno, int blkno, int *cylptr, int
 
    if ((blkno < 0) || (blkno >= currdisk->numblocks)) {
       fprintf(stderr, "Invalid blkno at simpledisk_get_mapping: %d\n", blkno);
-      exit(0);
+      exit(1);
    }
 
    if (cylptr) {
@@ -711,53 +702,13 @@ void simpledisk_cleanstats (void)
 
    for (i=0; i<MAXDEVICES; i++) {
       simpledisk_t *currdisk = getsimpledisk (i);
-      if (currdisk->inited) {
+      if (currdisk) {
          ioqueue_cleanstats(currdisk->queue);
       }
    }
 }
 
 
-void simpledisk_param_override (char *paramname, char *paramval, int first, int last)
-{
-   int i;
-
-   if (first == -1) {
-      first = 0;
-      last = MAXDEVICES - 1;
-   } else if ((first < 0) || (last >= MAXDEVICES) || (last < first)) {
-      fprintf(stderr, "Invalid range at simpledisk_param_override: %d - %d\n", first, last);
-      exit(0);
-   }
-
-   for (i=first; i<=last; i++) {
-      simpledisk_t *currdisk = getsimpledisk (i);
-      if (currdisk->inited == 0) {
-         continue;
-      }
-
-      if ((strcmp(paramname, "ioqueue_schedalg") == 0) ||
-	  (strcmp(paramname, "ioqueue_seqscheme") == 0) ||
-	  (strcmp(paramname, "ioqueue_cylmaptype") == 0) ||
-	  (strcmp(paramname, "ioqueue_to_time") == 0) ||
-	  (strcmp(paramname, "ioqueue_timeout_schedalg") == 0) ||
-	  (strcmp(paramname, "ioqueue_priority_schedalg") == 0) ||
-	  (strcmp(paramname, "ioqueue_priority_mix") == 0)) {
-         simpledisk_t *currdisk = getsimpledisk (i);
-         ioqueue_param_override(currdisk->queue, paramname, paramval);
-
-      } else if (strcmp(paramname, "maxqlen") == 0) {
-         if (sscanf(paramval, "%d\n", &currdisk->maxqlen) != 1) {
-            fprintf(stderr, "Error reading maxqlen in disk_param_override\n");
-            exit(0);
-         }
-
-      } else {
-	 fprintf(stderr, "Unsupported param to override at simpledisk_param_override: %s\n", paramname);
-	 exit(0);
-      }
-   }
-}
 
 
 static void simpledisk_statinit (int devno, int firsttime)
@@ -790,16 +741,17 @@ void simpledisk_setcallbacks ()
 
 static void simpledisk_initialize_diskinfo ()
 {
-   disksim->simplediskinfo = DISKSIM_malloc (sizeof(simplediskinfo_t));
-   bzero (disksim->simplediskinfo, sizeof(simplediskinfo_t));
-   disksim->simplediskinfo->simpledisks = DISKSIM_malloc(MAXDEVICES * (sizeof(simpledisk_t)));
-   bzero (disksim->simplediskinfo->simpledisks, (MAXDEVICES * (sizeof(simpledisk_t))));
+   disksim->simplediskinfo = malloc (sizeof(simplediskinfo_t));
+   bzero ((char *)disksim->simplediskinfo, sizeof(simplediskinfo_t));
+   disksim->simplediskinfo->simpledisks = malloc(MAXDEVICES * (sizeof(simpledisk_t)));
+   disksim->simplediskinfo->simpledisks_len = MAXDEVICES;
+   bzero ((char *)disksim->simplediskinfo->simpledisks, (MAXDEVICES * (sizeof(simpledisk_t))));
 }
 
 
 void simpledisk_initialize (void)
 {
-   int i, j;
+   int i;
 
    if (disksim->simplediskinfo == NULL) {
       simpledisk_initialize_diskinfo ();
@@ -812,20 +764,15 @@ fprintf (outputfile, "Entered simpledisk_initialize - numsimpledisks %d\n", nums
 
    for (i=0; i<MAXDEVICES; i++) {
       simpledisk_t *currdisk = getsimpledisk (i);
-      if (currdisk->inited) {
+      if(!currdisk) continue;
+/*        if (!currdisk->inited) { */
          currdisk->media_busy = FALSE;
          currdisk->reconnect_reason = -1;
          addlisttoextraq ((event **) &currdisk->buswait);
          currdisk->busowned = -1;
-         currdisk->numinbuses = 0;
-         for (j=0; j<MAXINBUSES; j++) {
-            currdisk->inbuses[j] = -1;
-            currdisk->depth[j] = -1;
-            currdisk->slotno[j] = -1;
-         }
          ioqueue_initialize (currdisk->queue, i);
          simpledisk_statinit(i, TRUE);
-      }
+/*        } */
    }
 }
 
@@ -836,7 +783,7 @@ void simpledisk_resetstats (void)
 
    for (i=0; i<MAXDEVICES; i++) {
       simpledisk_t *currdisk = getsimpledisk (i);
-      if (currdisk->inited) {
+      if (currdisk) {
          ioqueue_resetstats(currdisk->queue);
          simpledisk_statinit(i, 0);
       }
@@ -844,52 +791,72 @@ void simpledisk_resetstats (void)
 }
 
 
-void simpledisk_read_toprints (FILE *parfile)
-{
+
+int simpledisk_add(struct simpledisk *d) {
+  int c;
+
+  if(!disksim->simplediskinfo) simpledisk_initialize_diskinfo();
+  
+  for(c = 0; c < disksim->simplediskinfo->simpledisks_len; c++) {
+    if(!disksim->simplediskinfo->simpledisks[c]) {
+      disksim->simplediskinfo->simpledisks[c] = d;
+      numsimpledisks++;
+      return c;
+    }
+  }
+
+  /* note that numdisks must be equal to diskinfo->disks_len */
+  disksim->simplediskinfo->simpledisks = 
+    realloc(disksim->simplediskinfo->simpledisks, 
+	    2 * c * sizeof(struct simpledisk *));
+
+  bzero(disksim->simplediskinfo->simpledisks + numsimpledisks, 
+	numsimpledisks);
+
+  disksim->simplediskinfo->simpledisks[c] = d;
+  numsimpledisks++;
+  disksim->simplediskinfo->simpledisks_len *= 2;
+  return c;
 }
 
 
-void simpledisk_read_specs (FILE *parfile, int devno, int copies)
+struct simpledisk *disksim_simpledisk_loadparams(struct lp_block *b)
 {
-   int i;
-   simpledisk_t *currdisk;
+  /* temp vars for parameters */
+  struct simpledisk *result;
+  int num;
 
-   if (disksim->simplediskinfo == NULL) {
-      simpledisk_initialize_diskinfo();
-   }
+  if(!disksim->simplediskinfo) simpledisk_initialize_diskinfo();
 
-   numsimpledisks += copies;
+  result = malloc(sizeof(struct simpledisk));
+  if(!result) return 0;
+  bzero(result, sizeof(struct simpledisk));
+  
+  num = simpledisk_add(result);
 
-      currdisk = getsimpledisk (devno);
+  result->hdr = simpledisk_hdr_initializer; 
+  if(b->name)
+    result->hdr.device_name = strdup(b->name);
 
-      getparam_int(parfile, "Number of blocks", &currdisk->numblocks, 1, 0, 0);
-      getparam_double(parfile, "Access time (in msecs)", &currdisk->acctime, 1, (double) 0.0, (double) 0.0);
-      getparam_double(parfile, "Command overhead (in msecs)", &currdisk->overhead, 1, (double) 0.0, (double) 0.0);
-      getparam_double(parfile, "Bus transaction latency (in msecs)", &currdisk->bus_transaction_latency, 1, (double) 0.0, (double) 0.0);
-      getparam_double(parfile, "Bulk sector transfer time", &currdisk->blktranstime, 1, (double) 0.0, (double) 0.0);
-      getparam_bool(parfile, "Never disconnect:", &currdisk->neverdisconnect);
-      getparam_bool(parfile, "Print stats for simpledisk:", &currdisk->printstats);
-      getparam_int(parfile, "Max queue length at simpledisk", &currdisk->maxqlen, 1, 0, 0);
-      currdisk->queue = ioqueue_readparams(parfile, device_printqueuestats, device_printcritstats, device_printidlestats, device_printintarrstats, device_printsizestats);
+  //#include "modules/disksim_simpledisk_param.c"
+  lp_loadparams(result, b, &disksim_simpledisk_mod);
 
-      for (i=devno; i<(devno+copies); i++) {
-         simpledisk_t *currdisk2 = getsimpledisk (i);
-         currdisk2->inited = 1;
-         currdisk2->devno = i;
-         currdisk2->acctime = currdisk->acctime;
-         currdisk2->overhead = currdisk->overhead;
-         currdisk2->bus_transaction_latency = currdisk->bus_transaction_latency;
-         currdisk2->blktranstime = currdisk->blktranstime;
-         currdisk2->neverdisconnect = currdisk->neverdisconnect;
-         currdisk2->printstats = currdisk->printstats;
-         currdisk2->maxqlen = currdisk->maxqlen;
-	 if (i != devno) {
-	    currdisk2->queue = ioqueue_copy(currdisk->queue);
-	 }
-      }
+
+
+  device_add((struct device_header *)result, num);
+  return result;
 }
 
 
+struct simpledisk *simpledisk_copy(struct simpledisk *orig) {
+  struct simpledisk *result = malloc(sizeof(struct simpledisk));
+  bzero(result, sizeof(struct simpledisk));
+  memcpy(result, orig, sizeof(struct simpledisk));
+  result->queue = ioqueue_copy(orig->queue);
+
+  return result;
+}
+
 void simpledisk_set_syncset (int setstart, int setend)
 {
 }
@@ -968,7 +935,7 @@ void simpledisk_printstats (void)
    diskcnt = 0;
    for (i=0; i<MAXDEVICES; i++) {
       simpledisk_t *currdisk = getsimpledisk (i);
-      if (currdisk->inited) {
+      if (currdisk) {
          queueset[diskcnt] = currdisk->queue;
          reqcnt += ioqueue_get_number_of_requests(currdisk->queue);
          diskcnt++;
@@ -986,7 +953,7 @@ void simpledisk_printstats (void)
    diskcnt = 0;
    for (i=0; i<MAXDEVICES; i++) {
       simpledisk_t *currdisk = getsimpledisk (i);
-      if (currdisk->inited) {
+      if (currdisk) {
          set[diskcnt] = i;
          diskcnt++;
       }
@@ -1019,3 +986,38 @@ void simpledisk_printstats (void)
    }
 }
 
+
+double simpledisk_get_seektime (int devno, 
+				ioreq_event *req, 
+				int checkcache, 
+				double maxtime)
+{
+  fprintf(stderr, "device_get_seektime not supported for simpledisk devno %d\n",  devno);
+  assert(0);
+}
+
+/* default simpledisk dev header */
+struct device_header simpledisk_hdr_initializer = { 
+  DEVICETYPE_SIMPLEDISK,
+  sizeof(struct simpledisk),
+  "unnamed simpledisk",
+  (void *)simpledisk_copy,
+  simpledisk_set_depth,
+  simpledisk_get_depth,
+  simpledisk_get_inbus,
+  simpledisk_get_busno,
+  simpledisk_get_slotno,
+  simpledisk_get_number_of_blocks,
+  simpledisk_get_maxoutstanding,
+  simpledisk_get_numcyls,
+  simpledisk_get_blktranstime,
+  simpledisk_get_avg_sectpercyl,
+  simpledisk_get_mapping,
+  simpledisk_event_arrive,
+  simpledisk_get_distance,
+  simpledisk_get_servtime,
+  simpledisk_get_seektime,
+  simpledisk_get_acctime,
+  simpledisk_bus_delay_complete,
+  simpledisk_bus_ownership_grant
+};
